from .__init__ import ctypes, typing, abc, \
    SDL_POINTER, SDL_ENUM, SDL_FUNC, SDL_TYPE, SDL_BINARY

from .SDL_blendmode import SDL_BlendMode
from .SDL_properties import SDL_PropertiesID
from .SDL_pixels import SDL_PixelFormat, SDL_Colorspace, SDL_Palette
from .SDL_iostream import SDL_IOStream
from .SDL_rect import SDL_Rect

SDL_SurfaceFlags: typing.TypeAlias = SDL_TYPE["SDL_SurfaceFlags", ctypes.c_uint32]

SDL_SURFACE_PREALLOCATED, SDL_SURFACE_LOCK_NEEDED, SDL_SURFACE_LOCKED, \
    SDL_SURFACE_SIMD_ALIGNED = 0x00000001, 0x00000002, 0x00000004, 0x00000008

SDL_MUSTLOCK: abc.Callable[..., bool] = lambda s: \
    (s.flags & SDL_SURFACE_LOCK_NEEDED) == SDL_SURFACE_LOCK_NEEDED

SDL_ScaleMode: typing.TypeAlias = SDL_TYPE["SDL_ScaleMode", SDL_ENUM]

SDL_SCALEMODE_INVALID, SDL_SCALEMODE_NEAREST, SDL_SCALEMODE_LINEAR = range(-1, 2)

SDL_FlipMode: typing.TypeAlias = SDL_TYPE["SDL_FlipMode", SDL_ENUM]

SDL_FLIP_NONE, SDL_FLIP_HORIZONTAL, SDL_FLIP_VERTICAL = range(3)

class SDL_Surface(ctypes.Structure):
    _fields_ = [
        ("flags", SDL_SurfaceFlags),
        ("format", SDL_PixelFormat),
        ("w", ctypes.c_int),
        ("h", ctypes.c_int),
        ("pitch", ctypes.c_int),
        ("pixels", ctypes.c_void_p),
        ("refcount", ctypes.c_int),
        ("reserved", ctypes.c_void_p)
    ]

SDL_CreateSurface: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_CreateSurface", SDL_POINTER[SDL_Surface], [ctypes.c_int, ctypes.c_int, SDL_PixelFormat], SDL_BINARY]
SDL_CreateSurfaceFrom: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_CreateSurfaceFrom", SDL_POINTER[SDL_Surface], [ctypes.c_int, ctypes.c_int, SDL_PixelFormat, ctypes.c_void_p, ctypes.c_int], SDL_BINARY]
SDL_DestroySurface: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_DestroySurface", None, [SDL_POINTER[SDL_Surface]], SDL_BINARY]

SDL_GetSurfaceProperties: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetSurfaceProperties", SDL_PropertiesID, [SDL_POINTER[SDL_Surface]], SDL_BINARY]

SDL_PROP_SURFACE_SDR_WHITE_POINT_FLOAT: bytes = "SDL.surface.SDR_white_point".encode()
SDL_PROP_SURFACE_HDR_HEADROOM_FLOAT: bytes = "SDL.surface.HDR_headroom".encode()
SDL_PROP_SURFACE_TONEMAP_OPERATOR_STRING: bytes = "SDL.surface.tonemap".encode()
SDL_PROP_SURFACE_HOTSPOT_X_NUMBER: bytes = "SDL.surface.hotspot.x".encode()
SDL_PROP_SURFACE_HOTSPOT_Y_NUMBER: bytes = "SDL.surface.hotspot.y".encode()

SDL_SetSurfaceColorspace: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetSurfaceColorspace", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_Colorspace], SDL_BINARY]
SDL_GetSurfaceColorspace: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetSurfaceColorspace", SDL_Colorspace, [SDL_POINTER[SDL_Surface]], SDL_BINARY]

SDL_CreateSurfacePalette: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_CreateSurfacePalette", SDL_POINTER[SDL_Palette], [SDL_POINTER[SDL_Surface]], SDL_BINARY]
SDL_SetSurfacePalette: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetSurfacePalette", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Palette]], SDL_BINARY]
SDL_GetSurfacePalette: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetSurfacePalette", SDL_POINTER[SDL_Palette], [SDL_POINTER[SDL_Surface]], SDL_BINARY]
SDL_AddSurfaceAlternateImage: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_AddSurfaceAlternateImage", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Surface]], SDL_BINARY]
SDL_SurfaceHasAlternateImages: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SurfaceHasAlternateImages", ctypes.c_bool, [SDL_POINTER[SDL_Surface]], SDL_BINARY]
SDL_GetSurfaceImages: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetSurfaceImages", SDL_POINTER[SDL_POINTER[SDL_Surface]], [SDL_POINTER[SDL_Surface], SDL_POINTER[ctypes.c_int]], SDL_BINARY]
SDL_RemoveSurfaceAlternateImages: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_RemoveSurfaceAlternateImages", None, [SDL_POINTER[SDL_Surface]], SDL_BINARY]
SDL_LockSurface: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_LockSurface", ctypes.c_bool, [SDL_POINTER[SDL_Surface]], SDL_BINARY]
SDL_UnlockSurface: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_UnlockSurface", None, [SDL_POINTER[SDL_Surface]], SDL_BINARY]

SDL_LoadBMP_IO: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_LoadBMP_IO", SDL_POINTER[SDL_Surface], [SDL_POINTER[SDL_IOStream], ctypes.c_bool], SDL_BINARY]
SDL_LoadBMP: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_LoadBMP", SDL_POINTER[SDL_Surface], [ctypes.c_char_p], SDL_BINARY]

SDL_SaveBMP_IO: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SaveBMP_IO", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_IOStream], ctypes.c_bool], SDL_BINARY]
SDL_SaveBMP: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SaveBMP", ctypes.c_bool, [SDL_POINTER[SDL_Surface], ctypes.c_char_p], SDL_BINARY]

SDL_SetSurfaceRLE: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetSurfaceRLE", ctypes.c_bool, [SDL_POINTER[SDL_Surface], ctypes.c_bool], SDL_BINARY]
SDL_SurfaceHasRLE: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SurfaceHasRLE", ctypes.c_bool, [SDL_POINTER[SDL_Surface]], SDL_BINARY]

SDL_SetSurfaceColorKey: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetSurfaceColorKey", ctypes.c_bool, [SDL_POINTER[SDL_Surface], ctypes.c_bool, ctypes.c_uint32], SDL_BINARY]
SDL_SurfaceHasColorKey: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SurfaceHasColorKey", ctypes.c_bool, [SDL_POINTER[SDL_Surface]], SDL_BINARY]
SDL_GetSurfaceColorKey: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetSurfaceColorKey", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[ctypes.c_uint32]], SDL_BINARY]

SDL_SetSurfaceColorMod: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetSurfaceColorMod", ctypes.c_bool, [SDL_POINTER[SDL_Surface], ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8], SDL_BINARY]
SDL_GetSurfaceColorMod: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetSurfaceColorMod", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[ctypes.c_uint8], SDL_POINTER[ctypes.c_uint8], SDL_POINTER[ctypes.c_uint8]], SDL_BINARY]

SDL_SetSurfaceAlphaMod: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetSurfaceAlphaMod", ctypes.c_bool, [SDL_POINTER[SDL_Surface], ctypes.c_uint8], SDL_BINARY]
SDL_GetSurfaceAlphaMod: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetSurfaceAlphaMod", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[ctypes.c_uint8]], SDL_BINARY]

SDL_SetSurfaceBlendMode: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetSurfaceBlendMode", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_BlendMode], SDL_BINARY]
SDL_GetSurfaceBlendMode: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetSurfaceBlendMode", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_BlendMode]], SDL_BINARY]

SDL_SetSurfaceClipRect: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetSurfaceClipRect", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect]], SDL_BINARY]
SDL_GetSurfaceClipRect: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetSurfaceClipRect", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect]], SDL_BINARY]

SDL_FlipSurface: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_FlipSurface", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_FlipMode], SDL_BINARY]
SDL_DuplicateSurface: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_DuplicateSurface", SDL_POINTER[SDL_Surface], [SDL_POINTER[SDL_Surface]], SDL_BINARY]
SDL_ScaleSurface: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ScaleSurface", SDL_POINTER[SDL_Surface], [SDL_POINTER[SDL_Surface], ctypes.c_int, ctypes.c_int, SDL_ScaleMode], SDL_BINARY]

SDL_ConvertSurface: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ConvertSurface", SDL_POINTER[SDL_Surface], [SDL_POINTER[SDL_Surface], SDL_PixelFormat], SDL_BINARY]
SDL_ConvertSurfaceAndColorspace: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ConvertSurfaceAndColorspace", SDL_POINTER[SDL_Surface], [SDL_POINTER[SDL_Surface], SDL_PixelFormat, SDL_POINTER[SDL_Palette], SDL_Colorspace, SDL_PropertiesID], SDL_BINARY]
SDL_ConvertPixels: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ConvertPixels", ctypes.c_bool, [ctypes.c_int, ctypes.c_int, SDL_PixelFormat, ctypes.c_void_p, ctypes.c_int, SDL_PixelFormat, ctypes.c_void_p, ctypes.c_int], SDL_BINARY]
SDL_ConvertPixelsAndColorspace: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ConvertPixelsAndColorspace", ctypes.c_bool, [ctypes.c_int, ctypes.c_int, SDL_PixelFormat, SDL_Colorspace, SDL_PropertiesID, ctypes.c_void_p, ctypes.c_int, SDL_PixelFormat, SDL_Colorspace, SDL_PropertiesID, ctypes.c_void_p, ctypes.c_int], SDL_BINARY]

SDL_PremultiplyAlpha: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_PremultiplyAlpha", ctypes.c_bool, [ctypes.c_int, ctypes.c_int, SDL_PixelFormat, ctypes.c_void_p, ctypes.c_int, SDL_PixelFormat, ctypes.c_void_p, ctypes.c_int, ctypes.c_bool], SDL_BINARY]
SDL_PremultiplySurfaceAlpha: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_PremultiplySurfaceAlpha", ctypes.c_bool, [SDL_POINTER[SDL_Surface], ctypes.c_bool], SDL_BINARY]

SDL_ClearSurface: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ClearSurface", ctypes.c_bool, [SDL_POINTER[SDL_Surface], ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float], SDL_BINARY]

SDL_FillSurfaceRect: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_FillSurfaceRect", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect], ctypes.c_uint32], SDL_BINARY]
SDL_FillSurfaceRects: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_FillSurfaceRects", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect], ctypes.c_int, ctypes.c_uint32], SDL_BINARY]

SDL_BlitSurface: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_BlitSurface", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect], SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect]], SDL_BINARY]
SDL_BlitSurfaceUnchecked: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_BlitSurfaceUnchecked", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect], SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect]], SDL_BINARY]
SDL_BlitSurfaceScaled: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_BlitSurfaceScaled", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect], SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect], SDL_ScaleMode], SDL_BINARY]
SDL_BlitSurfaceUncheckedScaled: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_BlitSurfaceUncheckedScaled", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect], SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect], SDL_ScaleMode], SDL_BINARY]
SDL_StretchSurface: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_StretchSurface", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect], SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect], SDL_ScaleMode], SDL_BINARY]
SDL_BlitSurfaceTiled: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_BlitSurfaceTiled", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect], SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect]], SDL_BINARY]
SDL_BlitSurfaceTiledWithScale: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_BlitSurfaceTiledWithScale", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect], ctypes.c_float, SDL_ScaleMode, SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect]], SDL_BINARY]
SDL_BlitSurface9Grid: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_BlitSurface9Grid", ctypes.c_bool, [SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect], ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_float, SDL_ScaleMode, SDL_POINTER[SDL_Surface], SDL_POINTER[SDL_Rect]], SDL_BINARY]

SDL_MapSurfaceRGB: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_MapSurfaceRGB", ctypes.c_uint32, [SDL_POINTER[SDL_Surface], ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8], SDL_BINARY]
SDL_MapSurfaceRGBA: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_MapSurfaceRGBA", ctypes.c_uint32, [SDL_POINTER[SDL_Surface], ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8], SDL_BINARY]

SDL_ReadSurfacePixel: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ReadSurfacePixel", ctypes.c_bool, [SDL_POINTER[SDL_Surface], ctypes.c_int, ctypes.c_int, SDL_POINTER[ctypes.c_uint8], SDL_POINTER[ctypes.c_uint8], SDL_POINTER[ctypes.c_uint8], SDL_POINTER[ctypes.c_uint8]], SDL_BINARY]
SDL_ReadSurfacePixelFloat: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ReadSurfacePixelFloat", ctypes.c_bool, [SDL_POINTER[SDL_Surface], ctypes.c_int, ctypes.c_int, SDL_POINTER[ctypes.c_float], SDL_POINTER[ctypes.c_float], SDL_POINTER[ctypes.c_float], SDL_POINTER[ctypes.c_float]], SDL_BINARY]

SDL_WriteSurfacePixel: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_WriteSurfacePixel", ctypes.c_bool, [SDL_POINTER[SDL_Surface], ctypes.c_int, ctypes.c_int, ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8], SDL_BINARY]
SDL_WriteSurfacePixelFloat: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_WriteSurfacePixelFloat", ctypes.c_bool, [SDL_POINTER[SDL_Surface], ctypes.c_int, ctypes.c_int, ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float], SDL_BINARY]


from .__init__ import sys, array

SDL_LIL_ENDIAN, SDL_BIG_ENDIAN = 1234, 4321
SDL_BYTEORDER: int = {"little": SDL_LIL_ENDIAN, "big": SDL_BIG_ENDIAN}[sys.byteorder]

def SDL_Swap16(x: int) -> int:
    return ((x << 8) & 0xFF00) | ((x >> 8) & 0x00FF)

def SDL_Swap32(x: int) -> int:
    return ((x << 24) & 0xFF000000) | ((x << 8) & 0x00FF0000) | ((x >> 8) & 0x0000FF00) | ((x >> 24) & 0x000000FF)

def SDL_Swap64(x: int) -> int:
    return (SDL_Swap32(x & 0xFFFFFFFF) << 32) | (SDL_Swap32(x >> 32 & 0xFFFFFFFF))

def SDL_SwapFloat(x: float) -> float:
    y = array.array("d", (x, ))
    y.byteswap(); return y[0]

if SDL_BYTEORDER == SDL_LIL_ENDIAN:
    SDL_SwapBE16, SDL_SwapBE32, SDL_SwapBE64, SDL_SwapFloatBE = SDL_Swap16, SDL_Swap32, SDL_Swap64, SDL_SwapFloat
    SDL_SwapLE16 = SDL_SwapLE32 = SDL_SwapLE64 = SDL_SwapFloatLE = lambda x: x

else:
    SDL_SwapLE16, SDL_SwapLE32, SDL_SwapLE64, SDL_SwapFloatLE = SDL_Swap16, SDL_Swap32, SDL_Swap64, SDL_SwapFloat
    SDL_SwapBE16 = SDL_SwapBE32 = SDL_SwapBE64 = SDL_SwapFloatBE = lambda x: x


from .__init__ import ctypes, typing, abc, SDL_VA_LIST, \
    SDL_POINTER, SDL_FUNC_TYPE, SDL_FUNC, SDL_TYPE, SDL_BINARY

SDL_arraysize: abc.Callable[[typing.Any], int] = lambda array: \
    ctypes.sizeof(array) // ctypes.sizeof(array[0])

def SDL_FOURCC(a: int, b: int, c: int, d: int) -> int:
    return (ord(a) << 0) | (ord(b) << 8) | (ord(c) << 16) | (ord(d) << 24)

Sint8: typing.TypeAlias = SDL_TYPE["Sint8", ctypes.c_int8]

SDL_MAX_SINT8, SDL_MIN_SINT8 = 0x7F, ~0X7F

Uint8: typing.TypeAlias = SDL_TYPE["Uint8", ctypes.c_uint8]

SDL_MAX_UINT8, SDL_MIN_UINT8 = 0xFF, 0X00

Sint16: typing.TypeAlias = SDL_TYPE["Sint16", ctypes.c_int16]

SDL_MAX_SINT16, SDL_MIN_SINT16 = 0x7FFF, ~0X7FFF

Uint16: typing.TypeAlias = SDL_TYPE["Uint16", ctypes.c_uint16]

SDL_MAX_UINT16, SDL_MIN_UINT16 = 0xFFFF, 0X0000

Sint32: typing.TypeAlias = SDL_TYPE["Sint32", ctypes.c_int32]

SDL_MAX_SINT32, SDL_MIN_SINT32 = 0x7FFFFFFF, ~0X7FFFFFFF

Uint32: typing.TypeAlias = SDL_TYPE["Uint32", ctypes.c_uint32]

SDL_MAX_UINT32, SDL_MIN_UINT32 = 0xFFFFFFFF, 0X00000000

Sint64: typing.TypeAlias = SDL_TYPE["Sint64", ctypes.c_int64]

SDL_MAX_SINT64, SDL_MIN_SINT64 = 0x7FFFFFFFFFFFFFFF, ~0X7FFFFFFFFFFFFFFF

Uint64: typing.TypeAlias = SDL_TYPE["Uint64", ctypes.c_uint64]

SDL_MAX_UINT64, SDL_MIN_UINT64 = 0xFFFFFFFFFFFFFFFF, 0X0000000000000000

SDL_Time: typing.TypeAlias = SDL_TYPE["SDL_Time", Sint64]

SDL_MAX_TIME, SDL_MIN_TIME = SDL_MAX_SINT64, SDL_MIN_SINT64

SDL_FLT_EPSILON: float = 1.1920928955078125e-07

SDL_malloc: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_malloc", ctypes.c_void_p, [ctypes.c_size_t], SDL_BINARY]
SDL_calloc: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_calloc", ctypes.c_void_p, [ctypes.c_size_t, ctypes.c_size_t], SDL_BINARY]
SDL_realloc: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_realloc", ctypes.c_void_p, [ctypes.c_void_p, ctypes.c_size_t], SDL_BINARY]
SDL_free: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_free", None, [ctypes.c_void_p], SDL_BINARY]

SDL_malloc_func: typing.TypeAlias = SDL_FUNC_TYPE["SDL_malloc_func", ctypes.c_void_p, [ctypes.c_size_t]]
SDL_calloc_func: typing.TypeAlias = SDL_FUNC_TYPE["SDL_calloc_func", ctypes.c_void_p, [ctypes.c_size_t, ctypes.c_size_t]]
SDL_realloc_func: typing.TypeAlias = SDL_FUNC_TYPE["SDL_realloc_func", ctypes.c_void_p, [ctypes.c_void_p, ctypes.c_size_t]]
SDL_free_func: typing.TypeAlias = SDL_FUNC_TYPE["SDL_free_func", None, [ctypes.c_void_p]]

SDL_GetOriginalMemoryFunctions: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetOriginalMemoryFunctions", None, [SDL_POINTER[SDL_malloc_func], SDL_POINTER[SDL_calloc_func], SDL_POINTER[SDL_realloc_func], SDL_POINTER[SDL_free_func]], SDL_BINARY]
SDL_GetMemoryFunctions: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetMemoryFunctions", None, [SDL_POINTER[SDL_malloc_func], SDL_POINTER[SDL_calloc_func], SDL_POINTER[SDL_realloc_func], SDL_POINTER[SDL_free_func]], SDL_BINARY]
SDL_SetMemoryFunctions: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetMemoryFunctions", ctypes.c_bool, [SDL_malloc_func, SDL_calloc_func, SDL_realloc_func, SDL_free_func], SDL_BINARY]

SDL_aligned_alloc: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_aligned_alloc", ctypes.c_void_p, [ctypes.c_size_t, ctypes.c_size_t], SDL_BINARY]
SDL_aligned_free: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_aligned_free", None, [ctypes.c_void_p], SDL_BINARY]

SDL_GetNumAllocations: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetNumAllocations", ctypes.c_int, [], SDL_BINARY]

class SDL_Environment(ctypes.c_void_p):
    ...

SDL_GetEnvironment: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetEnvironment", SDL_POINTER[SDL_Environment], [], SDL_BINARY]
SDL_CreateEnvironment: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_CreateEnvironment", SDL_POINTER[SDL_Environment], [ctypes.c_bool], SDL_BINARY]
SDL_GetEnvironmentVariable: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetEnvironmentVariable", ctypes.c_char_p, [SDL_POINTER[SDL_Environment], ctypes.c_char_p], SDL_BINARY]
SDL_GetEnvironmentVariables: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetEnvironmentVariables", SDL_POINTER[ctypes.c_char_p], [SDL_POINTER[SDL_Environment]], SDL_BINARY]
SDL_SetEnvironmentVariable: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetEnvironmentVariable", ctypes.c_bool, [SDL_POINTER[SDL_Environment], ctypes.c_char_p, ctypes.c_char_p, ctypes.c_bool], SDL_BINARY]
SDL_UnsetEnvironmentVariable: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_UnsetEnvironmentVariable", ctypes.c_bool, [SDL_POINTER[SDL_Environment], ctypes.c_char_p], SDL_BINARY]
SDL_DestroyEnvironment: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_DestroyEnvironment", None, [SDL_POINTER[SDL_Environment]], SDL_BINARY]

SDL_getenv: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_getenv", ctypes.c_char_p, [ctypes.c_char_p], SDL_BINARY]
SDL_getenv_unsafe: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_getenv_unsafe", ctypes.c_char_p, [ctypes.c_char_p], SDL_BINARY]
SDL_setenv_unsafe: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_setenv_unsafe", ctypes.c_int, [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_int], SDL_BINARY]
SDL_unsetenv_unsafe: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_unsetenv_unsafe", ctypes.c_int, [ctypes.c_char_p], SDL_BINARY]

SDL_CompareCallback: typing.TypeAlias = SDL_FUNC_TYPE["SDL_CompareCallback", ctypes.c_int, [ctypes.c_void_p, ctypes.c_void_p]]

SDL_qsort: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_qsort", None, [ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t, SDL_CompareCallback], SDL_BINARY]
SDL_bsearch: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_bsearch", ctypes.c_void_p, [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t, SDL_CompareCallback], SDL_BINARY]

SDL_CompareCallback_r: typing.TypeAlias = SDL_FUNC_TYPE["SDL_CompareCallback_r", ctypes.c_int, [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p]]

SDL_qsort_r: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_qsort_r", None, [ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t, SDL_CompareCallback_r, ctypes.c_void_p], SDL_BINARY]
SDL_bsearch_r: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_bsearch_r", ctypes.c_void_p, [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t, SDL_CompareCallback_r, ctypes.c_void_p], SDL_BINARY]

SDL_abs: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_abs", ctypes.c_int, [ctypes.c_int], SDL_BINARY]

SDL_min: abc.Callable[[int | float, int | float], int | float] = lambda x, y: x if x < y else y
SDL_max: abc.Callable[[int | float, int | float], int | float] = lambda x, y: x if x > y else y
SDL_clamp: abc.Callable[[int | float, int | float, int | float], int | float] = lambda x, a, b: a if x < a else (b if x > b else x)

SDL_isalpha: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_isalpha", ctypes.c_int, [ctypes.c_int], SDL_BINARY]
SDL_isalnum: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_isalnum", ctypes.c_int, [ctypes.c_int], SDL_BINARY]
SDL_isblank: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_isblank", ctypes.c_int, [ctypes.c_int], SDL_BINARY]
SDL_iscntrl: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_iscntrl", ctypes.c_int, [ctypes.c_int], SDL_BINARY]
SDL_isdigit: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_isdigit", ctypes.c_int, [ctypes.c_int], SDL_BINARY]
SDL_isxdigit: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_isxdigit", ctypes.c_int, [ctypes.c_int], SDL_BINARY]
SDL_ispunct: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ispunct", ctypes.c_int, [ctypes.c_int], SDL_BINARY]
SDL_isspace: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_isspace", ctypes.c_int, [ctypes.c_int], SDL_BINARY]
SDL_isupper: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_isupper", ctypes.c_int, [ctypes.c_int], SDL_BINARY]
SDL_islower: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_islower", ctypes.c_int, [ctypes.c_int], SDL_BINARY]
SDL_isprint: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_isprint", ctypes.c_int, [ctypes.c_int], SDL_BINARY]
SDL_isgraph: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_isgraph", ctypes.c_int, [ctypes.c_int], SDL_BINARY]
SDL_toupper: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_toupper", ctypes.c_int, [ctypes.c_int], SDL_BINARY]
SDL_tolower: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_tolower", ctypes.c_int, [ctypes.c_int], SDL_BINARY]

SDL_crc16: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_crc16", ctypes.c_uint16, [ctypes.c_uint16, ctypes.c_void_p, ctypes.c_size_t], SDL_BINARY]
SDL_crc32: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_crc32", ctypes.c_uint32, [ctypes.c_uint32, ctypes.c_void_p, ctypes.c_size_t], SDL_BINARY]
SDL_murmur3_32: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_murmur3_32", ctypes.c_uint32, [ctypes.c_void_p, ctypes.c_size_t, ctypes.c_uint32], SDL_BINARY]

SDL_memcpy: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_memcpy", ctypes.c_void_p, [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t], SDL_BINARY]
SDL_memmove: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_memmove", ctypes.c_void_p, [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t], SDL_BINARY]
SDL_memset: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_memset", ctypes.c_void_p, [ctypes.c_void_p, ctypes.c_int, ctypes.c_size_t], SDL_BINARY]
SDL_memset4: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_memset4", ctypes.c_void_p, [ctypes.c_void_p, ctypes.c_uint32, ctypes.c_size_t], SDL_BINARY]

SDL_zero: abc.Callable[[ctypes.c_void_p], ctypes.c_void_p] = lambda x: SDL_memset(ctypes.byref(x), 0, ctypes.sizeof(x))
SDL_zerop: abc.Callable[[ctypes.c_void_p], ctypes.c_void_p] = lambda x: SDL_memset(x, 0, ctypes.sizeof(x.contents))
SDL_zeroa: abc.Callable[[ctypes.c_void_p], ctypes.c_void_p] = lambda x: SDL_memset(x, 0, ctypes.sizeof(x))

SDL_memcmp: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_memcmp", ctypes.c_int, [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t], SDL_BINARY]

SDL_wcslen: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_wcslen", ctypes.c_size_t, [ctypes.c_wchar_p], SDL_BINARY]
SDL_wcsnlen: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_wcsnlen", ctypes.c_size_t, [ctypes.c_wchar_p, ctypes.c_size_t], SDL_BINARY]
SDL_wcslcpy: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_wcslcpy", ctypes.c_size_t, [ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_size_t], SDL_BINARY]
SDL_wcslcat: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_wcslcat", ctypes.c_size_t, [ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_size_t], SDL_BINARY]
SDL_wcsdup: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_wcsdup", ctypes.c_wchar_p, [ctypes.c_wchar_p], SDL_BINARY]
SDL_wcsstr: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_wcsstr", ctypes.c_wchar_p, [ctypes.c_wchar_p, ctypes.c_wchar_p], SDL_BINARY]
SDL_wcsnstr: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_wcsnstr", ctypes.c_wchar_p, [ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_size_t], SDL_BINARY]
SDL_wcscmp: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_wcscmp", ctypes.c_int, [ctypes.c_wchar_p, ctypes.c_wchar_p], SDL_BINARY]
SDL_wcsncmp: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_wcsncmp", ctypes.c_int, [ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_size_t], SDL_BINARY]
SDL_wcscasecmp: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_wcscasecmp", ctypes.c_int, [ctypes.c_wchar_p, ctypes.c_wchar_p], SDL_BINARY]
SDL_wcsncasecmp: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_wcsncasecmp", ctypes.c_int, [ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_size_t], SDL_BINARY]
SDL_wcstol: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_wcstol", ctypes.c_long, [ctypes.c_wchar_p, SDL_POINTER[ctypes.c_wchar_p], ctypes.c_int], SDL_BINARY]

SDL_strlen: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strlen", ctypes.c_size_t, [ctypes.c_char_p], SDL_BINARY]
SDL_strnlen: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strnlen", ctypes.c_size_t, [ctypes.c_char_p, ctypes.c_size_t], SDL_BINARY]
SDL_strlcpy: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strlcpy", ctypes.c_size_t, [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_size_t], SDL_BINARY]
SDL_utf8strlcpy: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_utf8strlcpy", ctypes.c_size_t, [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_size_t], SDL_BINARY]
SDL_strlcat: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strlcat", ctypes.c_size_t, [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_size_t], SDL_BINARY]
SDL_strdup: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strdup", ctypes.c_char_p, [ctypes.c_char_p], SDL_BINARY]
SDL_strndup: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strndup", ctypes.c_char_p, [ctypes.c_char_p, ctypes.c_size_t], SDL_BINARY]
SDL_strrev: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strrev", ctypes.c_char_p, [ctypes.c_char_p], SDL_BINARY]
SDL_strupr: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strupr", ctypes.c_char_p, [ctypes.c_char_p], SDL_BINARY]
SDL_strlwr: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strlwr", ctypes.c_char_p, [ctypes.c_char_p], SDL_BINARY]
SDL_strchr: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strchr", ctypes.c_char_p, [ctypes.c_char_p, ctypes.c_int], SDL_BINARY]
SDL_strrchr: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strrchr", ctypes.c_char_p, [ctypes.c_char_p, ctypes.c_int], SDL_BINARY]
SDL_strstr: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strstr", ctypes.c_char_p, [ctypes.c_char_p, ctypes.c_char_p], SDL_BINARY]
SDL_strnstr: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strnstr", ctypes.c_char_p, [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_size_t], SDL_BINARY]
SDL_strcasestr: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strcasestr", ctypes.c_char_p, [ctypes.c_char_p, ctypes.c_char_p], SDL_BINARY]
SDL_strtok_r: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strtok_r", ctypes.c_char_p, [ctypes.c_char_p, ctypes.c_char_p, SDL_POINTER[ctypes.c_char_p]], SDL_BINARY]
SDL_utf8strlen: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_utf8strlen", ctypes.c_size_t, [ctypes.c_char_p], SDL_BINARY]
SDL_utf8strnlen: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_utf8strnlen", ctypes.c_size_t, [ctypes.c_char_p, ctypes.c_size_t], SDL_BINARY]

SDL_itoa: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_itoa", ctypes.c_char_p, [ctypes.c_int, ctypes.c_char_p, ctypes.c_int], SDL_BINARY]
SDL_uitoa: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_uitoa", ctypes.c_char_p, [ctypes.c_uint, ctypes.c_char_p, ctypes.c_int], SDL_BINARY]
SDL_ltoa: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ltoa", ctypes.c_char_p, [ctypes.c_long, ctypes.c_char_p, ctypes.c_int], SDL_BINARY]
SDL_ultoa: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ultoa", ctypes.c_char_p, [ctypes.c_ulong, ctypes.c_char_p, ctypes.c_int], SDL_BINARY]
SDL_lltoa: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_lltoa", ctypes.c_char_p, [ctypes.c_longlong, ctypes.c_char_p, ctypes.c_int], SDL_BINARY]
SDL_ulltoa: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ulltoa", ctypes.c_char_p, [ctypes.c_ulonglong, ctypes.c_char_p, ctypes.c_int], SDL_BINARY]

SDL_atoi: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_atoi", ctypes.c_int, [ctypes.c_char_p], SDL_BINARY]
SDL_atof: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_atof", ctypes.c_double, [ctypes.c_char_p], SDL_BINARY]
SDL_strtol: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strtol", ctypes.c_long, [ctypes.c_char_p, SDL_POINTER[ctypes.c_char_p], ctypes.c_int], SDL_BINARY]
SDL_strtoul: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strtoul", ctypes.c_ulong, [ctypes.c_char_p, SDL_POINTER[ctypes.c_char_p], ctypes.c_int], SDL_BINARY]
SDL_strtoll: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strtoll", ctypes.c_longlong, [ctypes.c_char_p, SDL_POINTER[ctypes.c_char_p], ctypes.c_int], SDL_BINARY]
SDL_strtoull: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strtoull", ctypes.c_ulonglong, [ctypes.c_char_p, SDL_POINTER[ctypes.c_char_p], ctypes.c_int], SDL_BINARY]
SDL_strtod: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strtod", ctypes.c_double, [ctypes.c_char_p, SDL_POINTER[ctypes.c_char_p]], SDL_BINARY]

SDL_strcmp: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strcmp", ctypes.c_int, [ctypes.c_char_p, ctypes.c_char_p], SDL_BINARY]
SDL_strncmp: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strncmp", ctypes.c_int, [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_size_t], SDL_BINARY]
SDL_strcasecmp: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strcasecmp", ctypes.c_int, [ctypes.c_char_p, ctypes.c_char_p], SDL_BINARY]
SDL_strncasecmp: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_strncasecmp", ctypes.c_int, [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_size_t], SDL_BINARY]

SDL_INVALID_UNICODE_CODEPOINT: int = 0xFFFD

SDL_StepUTF8: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_StepUTF8", ctypes.c_uint32, [SDL_POINTER[ctypes.c_char_p], SDL_POINTER[ctypes.c_size_t]], SDL_BINARY]
SDL_StepBackUTF8: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_StepBackUTF8", ctypes.c_uint32, [ctypes.c_char_p, SDL_POINTER[ctypes.c_char_p]], SDL_BINARY]
SDL_UCS4ToUTF8: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_UCS4ToUTF8", ctypes.c_char_p, [ctypes.c_uint32, ctypes.c_char_p], SDL_BINARY]

SDL_sscanf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_sscanf", ctypes.c_int, [ctypes.c_char_p, ctypes.c_char_p, ...], SDL_BINARY]
SDL_vsscanf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_vsscanf", ctypes.c_int, [ctypes.c_char_p, ctypes.c_char_p, SDL_VA_LIST], SDL_BINARY]
SDL_snprintf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_snprintf", ctypes.c_int, [ctypes.c_char_p, ctypes.c_size_t, ctypes.c_char_p, ...], SDL_BINARY]
SDL_swprintf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_swprintf", ctypes.c_int, [ctypes.c_wchar_p, ctypes.c_size_t, ctypes.c_wchar_p, ...], SDL_BINARY]
SDL_vsnprintf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_vsnprintf", ctypes.c_int, [ctypes.c_char_p, ctypes.c_size_t, ctypes.c_char_p, SDL_VA_LIST], SDL_BINARY]
SDL_vswprintf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_vswprintf", ctypes.c_int, [ctypes.c_wchar_p, ctypes.c_size_t, ctypes.c_wchar_p, SDL_VA_LIST], SDL_BINARY]
SDL_asprintf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_asprintf", ctypes.c_int, [SDL_POINTER[ctypes.c_char_p], ctypes.c_char_p, ...], SDL_BINARY]
SDL_vasprintf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_vasprintf", ctypes.c_int, [SDL_POINTER[ctypes.c_char_p], ctypes.c_char_p, SDL_VA_LIST], SDL_BINARY]

SDL_srand: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_srand", None, [ctypes.c_uint64], SDL_BINARY]
SDL_rand: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_rand", ctypes.c_int32, [ctypes.c_int32], SDL_BINARY]
SDL_randf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_randf", ctypes.c_float, [], SDL_BINARY]
SDL_rand_bits: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_rand_bits", ctypes.c_uint32, [], SDL_BINARY]
SDL_rand_r: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_rand_r", ctypes.c_int32, [SDL_POINTER[ctypes.c_uint64], ctypes.c_int32], SDL_BINARY]
SDL_randf_r: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_randf_r", ctypes.c_float, [SDL_POINTER[ctypes.c_uint64]], SDL_BINARY]
SDL_rand_bits_r: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_rand_bits_r", ctypes.c_uint32, [SDL_POINTER[ctypes.c_uint64]], SDL_BINARY]

SDL_PI_D = SDL_PI_F = 3.141592653589793238462643383279502884

SDL_acos: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_acos", ctypes.c_double, [ctypes.c_double], SDL_BINARY]
SDL_acosf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_acosf", ctypes.c_float, [ctypes.c_float], SDL_BINARY]
SDL_asin: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_asin", ctypes.c_double, [ctypes.c_double], SDL_BINARY]
SDL_asinf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_asinf", ctypes.c_float, [ctypes.c_float], SDL_BINARY]
SDL_atan: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_atan", ctypes.c_double, [ctypes.c_double], SDL_BINARY]
SDL_atanf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_atanf", ctypes.c_float, [ctypes.c_float], SDL_BINARY]
SDL_atan2: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_atan2", ctypes.c_double, [ctypes.c_double, ctypes.c_double], SDL_BINARY]
SDL_atan2f: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_atan2f", ctypes.c_float, [ctypes.c_float, ctypes.c_float], SDL_BINARY]
SDL_ceil: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ceil", ctypes.c_double, [ctypes.c_double], SDL_BINARY]
SDL_ceilf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ceilf", ctypes.c_float, [ctypes.c_float], SDL_BINARY]
SDL_copysign: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_copysign", ctypes.c_double, [ctypes.c_double, ctypes.c_double], SDL_BINARY]
SDL_copysignf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_copysignf", ctypes.c_float, [ctypes.c_float, ctypes.c_float], SDL_BINARY]
SDL_cos: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_cos", ctypes.c_double, [ctypes.c_double], SDL_BINARY]
SDL_cosf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_cosf", ctypes.c_float, [ctypes.c_float], SDL_BINARY]
SDL_exp: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_exp", ctypes.c_double, [ctypes.c_double], SDL_BINARY]
SDL_expf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_expf", ctypes.c_float, [ctypes.c_float], SDL_BINARY]
SDL_fabs: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_fabs", ctypes.c_double, [ctypes.c_double], SDL_BINARY]
SDL_fabsf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_fabsf", ctypes.c_float, [ctypes.c_float], SDL_BINARY]
SDL_floor: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_floor", ctypes.c_double, [ctypes.c_double], SDL_BINARY]
SDL_floorf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_floorf", ctypes.c_float, [ctypes.c_float], SDL_BINARY]
SDL_trunc: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_trunc", ctypes.c_double, [ctypes.c_double], SDL_BINARY]
SDL_truncf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_truncf", ctypes.c_float, [ctypes.c_float], SDL_BINARY]
SDL_fmod: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_fmod", ctypes.c_double, [ctypes.c_double, ctypes.c_double], SDL_BINARY]
SDL_fmodf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_fmodf", ctypes.c_float, [ctypes.c_float, ctypes.c_float], SDL_BINARY]
SDL_isinf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_isinf", ctypes.c_int, [ctypes.c_double], SDL_BINARY]
SDL_isinff: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_isinff", ctypes.c_int, [ctypes.c_float], SDL_BINARY]
SDL_isnan: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_isnan", ctypes.c_int, [ctypes.c_double], SDL_BINARY]
SDL_isnanf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_isnanf", ctypes.c_int, [ctypes.c_float], SDL_BINARY]
SDL_log: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_log", ctypes.c_double, [ctypes.c_double], SDL_BINARY]
SDL_logf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_logf", ctypes.c_float, [ctypes.c_float], SDL_BINARY]
SDL_log10: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_log10", ctypes.c_double, [ctypes.c_double], SDL_BINARY]
SDL_log10f: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_log10f", ctypes.c_float, [ctypes.c_float], SDL_BINARY]
SDL_modf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_modf", ctypes.c_double, [ctypes.c_double, SDL_POINTER[ctypes.c_double]], SDL_BINARY]
SDL_modff: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_modff", ctypes.c_float, [ctypes.c_float, SDL_POINTER[ctypes.c_float]], SDL_BINARY]
SDL_pow: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_pow", ctypes.c_double, [ctypes.c_double, ctypes.c_double], SDL_BINARY]
SDL_powf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_powf", ctypes.c_float, [ctypes.c_float, ctypes.c_float], SDL_BINARY]
SDL_round: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_round", ctypes.c_double, [ctypes.c_double], SDL_BINARY]
SDL_roundf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_roundf", ctypes.c_float, [ctypes.c_float], SDL_BINARY]
SDL_lround: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_lround", ctypes.c_long, [ctypes.c_double], SDL_BINARY]
SDL_lroundf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_lroundf", ctypes.c_long, [ctypes.c_float], SDL_BINARY]
SDL_scalbn: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_scalbn", ctypes.c_double, [ctypes.c_double, ctypes.c_int], SDL_BINARY]
SDL_scalbnf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_scalbnf", ctypes.c_float, [ctypes.c_float, ctypes.c_int], SDL_BINARY]
SDL_sin: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_sin", ctypes.c_double, [ctypes.c_double], SDL_BINARY]
SDL_sinf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_sinf", ctypes.c_float, [ctypes.c_float], SDL_BINARY]
SDL_sqrt: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_sqrt", ctypes.c_double, [ctypes.c_double], SDL_BINARY]
SDL_sqrtf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_sqrtf", ctypes.c_float, [ctypes.c_float], SDL_BINARY]
SDL_tan: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_tan", ctypes.c_double, [ctypes.c_double], SDL_BINARY]
SDL_tanf: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_tanf", ctypes.c_float, [ctypes.c_float], SDL_BINARY]

class SDL_iconv_t(ctypes.c_void_p):
    ...

SDL_iconv_open: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_iconv_open", SDL_iconv_t, [ctypes.c_char_p, ctypes.c_char_p], SDL_BINARY]
SDL_iconv_close: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_iconv_close", ctypes.c_int, [SDL_iconv_t], SDL_BINARY]
SDL_iconv: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_iconv", ctypes.c_size_t, [SDL_iconv_t, SDL_POINTER[ctypes.c_char_p], SDL_POINTER[ctypes.c_size_t], SDL_POINTER[ctypes.c_char_p], SDL_POINTER[ctypes.c_size_t]], SDL_BINARY]

SDL_ICONV_ERROR, SDL_ICONV_E2BIG, SDL_ICONV_EILSEQ, SDL_ICONV_EINVAL = -1, -2, -3, -4

SDL_iconv_string: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_iconv_string", ctypes.c_char_p, [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_size_t], SDL_BINARY]

SDL_iconv_utf8_locale: abc.Callable[[ctypes.c_char_p], ctypes.c_char_p] = lambda s: \
    SDL_iconv_string("".encode(), "UTF-8".encode(), s, SDL_strlen(s) + 1)

SDL_iconv_utf8_ucs2: abc.Callable[[ctypes.c_char_p], ctypes.c_char_p] = lambda s: \
    SDL_iconv_string("UCS-2".encode(), "UTF-8".encode(), s, SDL_strlen(s) + 1)

SDL_iconv_utf8_ucs4: abc.Callable[[ctypes.c_char_p], ctypes.c_char_p] = lambda s: \
    SDL_iconv_string("UCS-4".encode(), "UTF-8".encode(), s, SDL_strlen(s) + 1)

SDL_iconv_wchar_utf8: abc.Callable[[ctypes.c_char_p], ctypes.c_char_p] = lambda s: \
    SDL_iconv_string("UTF-8".encode(), "WCHAR_T".encode(), s, (SDL_wcslen(s) + 1) * ctypes.sizeof(ctypes.c_wchar))

SDL_FunctionPointer: typing.TypeAlias = SDL_FUNC_TYPE["SDL_FunctionPointer", None, []]


from .__init__ import ctypes, typing, abc, SDL_VA_LIST, SDL_FUNC, SDL_BINARY

SDL_SetError: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetError", ctypes.c_bool, [ctypes.c_char_p, ...], SDL_BINARY]
SDL_SetErrorV: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetErrorV", ctypes.c_bool, [ctypes.c_char_p, SDL_VA_LIST], SDL_BINARY]
SDL_OutOfMemory: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_OutOfMemory", ctypes.c_bool, [], SDL_BINARY]
SDL_GetError: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetError", ctypes.c_char_p, [], SDL_BINARY]
SDL_ClearError: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ClearError", ctypes.c_bool, [], SDL_BINARY]

def SDL_Unsupported() -> ctypes.c_int:
    return SDL_SetError("That operation is not supported".encode())

def SDL_InvalidParamError(param: ctypes.c_char_p) -> ctypes.c_int:
    return SDL_SetError("Parameter '%s' is invalid".encode(), param)


from .__init__ import ctypes, typing, abc, \
    SDL_POINTER, SDL_ENUM, SDL_FUNC, SDL_TYPE, SDL_BINARY

from .SDL_mouse import SDL_MouseID

SDL_TouchID: typing.TypeAlias = SDL_TYPE["SDL_TouchID", ctypes.c_uint64]
SDL_FingerID: typing.TypeAlias = SDL_TYPE["SDL_FingerID", ctypes.c_uint64]

SDL_TouchDeviceType: typing.TypeAlias = SDL_TYPE["SDL_TouchDeviceType", SDL_ENUM]

SDL_TOUCH_DEVICE_INVALID, SDL_TOUCH_DEVICE_DIRECT, SDL_TOUCH_DEVICE_INDIRECT_ABSOLUTE, \
    SDL_TOUCH_DEVICE_INDIRECT_RELATIVE = range(-1, 3)

class SDL_Finger(ctypes.Structure):
    _fields_ = [
        ("id", SDL_FingerID),
        ("x", ctypes.c_float),
        ("y", ctypes.c_float),
        ("pressure", ctypes.c_float)
    ]

SDL_TOUCH_MOUSEID, SDL_MOUSE_TOUCHID = \
    SDL_MouseID(-1), SDL_TouchID(-1)

SDL_GetTouchDevices: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetTouchDevices", SDL_POINTER[SDL_TouchID], [SDL_POINTER[ctypes.c_int]], SDL_BINARY]
SDL_GetTouchDeviceName: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetTouchDeviceName", ctypes.c_char_p, [SDL_TouchID], SDL_BINARY]
SDL_GetTouchDeviceType: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetTouchDeviceType", SDL_TouchDeviceType, [SDL_TouchID], SDL_BINARY]
SDL_GetTouchFingers: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetTouchFingers", SDL_POINTER[SDL_POINTER[SDL_Finger]], [SDL_TouchID, SDL_POINTER[ctypes.c_int]], SDL_BINARY]



from .__init__ import ctypes, typing, abc, SDL_FUNC, SDL_BINARY

SDL_MAJOR_VERSION, SDL_MINOR_VERSION, SDL_MICRO_VERSION = 3, 2, 16

SDL_VERSIONNUM: abc.Callable[[int, int, int], int] = \
    lambda major, minor, patch: major * 1000000 + minor * 1000 + patch

SDL_VERSIONNUM_MAJOR: abc.Callable[[int], int] = lambda version: int(version / 1000000)
SDL_VERSIONNUM_MINOR: abc.Callable[[int], int] = lambda version: int(version / 1000) % 1000
SDL_VERSIONNUM_MICRO: abc.Callable[[int], int] = lambda version: version % 1000

SDL_VERSION: int = SDL_VERSIONNUM(SDL_MAJOR_VERSION, SDL_MINOR_VERSION, SDL_MICRO_VERSION)
SDL_VERSION_ATLEAST: abc.Callable[[int, int, int], bool] = lambda x, y, z: SDL_VERSION >= SDL_VERSIONNUM(x, y, z)

SDL_GetVersion: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetVersion", ctypes.c_int, [], SDL_BINARY]
SDL_GetRevision: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetRevision", ctypes.c_char_p, [], SDL_BINARY]


from .__init__ import ctypes, typing, abc, \
    SDL_POINTER, SDL_FUNC, SDL_BINARY

class SDL_Locale(ctypes.Structure):
    _fields_ = [
        ("language", ctypes.c_char_p),
        ("country", ctypes.c_char_p)
    ]

SDL_GetPreferredLocales: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetPreferredLocales", SDL_POINTER[SDL_POINTER[SDL_Locale]], [SDL_POINTER[ctypes.c_int]], SDL_BINARY]


from .__init__ import ctypes, typing, abc, SDL_PLATFORM_SPECIFIC, \
    SDL_POINTER, SDL_FUNC, SDL_FUNC_TYPE, SDL_TYPE, SDL_BINARY, SDL_ENUM

from .SDL_video import SDL_DisplayID

if SDL_PLATFORM_SPECIFIC(system = ["Windows"]):
    import ctypes.wintypes as wintypes

    MSG: typing.TypeAlias = wintypes.tagMSG
    SDL_WindowsMessageHook: typing.TypeAlias = SDL_FUNC_TYPE["SDL_WindowsMessageHook", ctypes.c_bool, [ctypes.c_void_p, SDL_POINTER[MSG]]]
    SDL_SetWindowsMessageHook: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetWindowsMessageHook", None, [SDL_WindowsMessageHook, ctypes.c_void_p], SDL_BINARY]

    SDL_GetDirect3D9AdapterIndex: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetDirect3D9AdapterIndex", ctypes.c_int, [SDL_DisplayID], SDL_BINARY]
    SDL_GetDXGIOutputInfo: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetDXGIOutputInfo", ctypes.c_bool, [SDL_DisplayID, SDL_POINTER[ctypes.c_int], SDL_POINTER[ctypes.c_int]], SDL_BINARY]

    class XEvent(ctypes.Union):
        ...

    SDL_X11EventHook: typing.TypeAlias = SDL_FUNC_TYPE["SDL_X11EventHook", ctypes.c_bool, [ctypes.c_void_p, SDL_POINTER[XEvent]]]
    SDL_SetX11EventHook: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetX11EventHook", None, [SDL_X11EventHook, ctypes.c_void_p], SDL_BINARY] 

if SDL_PLATFORM_SPECIFIC(system = ["Linux"]):
    SDL_SetLinuxThreadPriority: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetLinuxThreadPriority", ctypes.c_bool, [ctypes.c_int64, ctypes.c_int], SDL_BINARY]
    SDL_SetLinuxThreadPriorityAndPolicy: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetLinuxThreadPriorityAndPolicy", ctypes.c_bool, [ctypes.c_int64, ctypes.c_int, ctypes.c_int], SDL_BINARY]

SDL_IsTablet: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_IsTablet", ctypes.c_bool, [], SDL_BINARY]
SDL_IsTV: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_IsTV", ctypes.c_bool, [], SDL_BINARY]

SDL_Sandbox: typing.TypeAlias = SDL_TYPE["SDL_Sandbox", SDL_ENUM]

SDL_SANDBOX_NONE, SDL_SANDBOX_UNKNOWN_CONTAINER, SDL_SANDBOX_FLATPAK, \
    SDL_SANDBOX_SNAP, SDL_SANDBOX_MACOS = range(5)

SDL_GetSandbox: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetSandbox", SDL_Sandbox, [], SDL_BINARY]

SDL_OnApplicationWillTerminate: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_OnApplicationWillTerminate", None, [], SDL_BINARY]
SDL_OnApplicationDidReceiveMemoryWarning: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_OnApplicationDidReceiveMemoryWarning", None, [], SDL_BINARY]
SDL_OnApplicationWillEnterBackground: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_OnApplicationWillEnterBackground", None, [], SDL_BINARY]
SDL_OnApplicationDidEnterBackground: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_OnApplicationDidEnterBackground", None, [], SDL_BINARY]
SDL_OnApplicationWillEnterForeground: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_OnApplicationWillEnterForeground", None, [], SDL_BINARY]
SDL_OnApplicationDidEnterForeground: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_OnApplicationDidEnterForeground", None, [], SDL_BINARY]


from .__init__ import ctypes, typing, abc, \
    SDL_FUNC_TYPE, SDL_FUNC, SDL_TYPE, SDL_BINARY

SDL_MS_PER_SECOND: int = 1000
SDL_US_PER_SECOND: int = 1000000
SDL_NS_PER_SECOND: int = 1000000000
SDL_NS_PER_MS: int = 1000000
SDL_NS_PER_US: int = 1000

SDL_SECONDS_TO_NS: abc.Callable[[int | float], int | float] = lambda s: s * SDL_NS_PER_SECOND
SDL_NS_TO_SECONDS: abc.Callable[[int | float], float] = lambda ns: ns / SDL_NS_PER_SECOND
SDL_MS_TO_NS: abc.Callable[[int | float], int | float] = lambda ms: ms * SDL_NS_PER_MS
SDL_NS_TO_MS: abc.Callable[[int | float], float] = lambda ns: ns / SDL_NS_PER_MS
SDL_US_TO_NS: abc.Callable[[int | float], int | float] = lambda us: us * SDL_NS_PER_US
SDL_NS_TO_US: abc.Callable[[int | float], float] = lambda ns: ns / SDL_NS_PER_US

SDL_GetTicks: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetTicks", ctypes.c_uint64, [], SDL_BINARY]
SDL_GetTicksNS: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetTicksNS", ctypes.c_uint64, [], SDL_BINARY]
SDL_GetPerformanceCounter: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetPerformanceCounter", ctypes.c_uint64, [], SDL_BINARY]
SDL_GetPerformanceFrequency: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetPerformanceFrequency", ctypes.c_uint64, [], SDL_BINARY]
SDL_Delay: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_Delay", None, [ctypes.c_uint32], SDL_BINARY]
SDL_DelayNS: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_DelayNS", None, [ctypes.c_uint64], SDL_BINARY]
SDL_DelayPrecise: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_DelayPrecise", None, [ctypes.c_uint64], SDL_BINARY]

SDL_TimerID: typing.TypeAlias = SDL_TYPE["SDL_TimerID", ctypes.c_uint32]
SDL_TimerCallback: typing.TypeAlias = SDL_FUNC_TYPE["SDL_TimerCallback", ctypes.c_uint32, [ctypes.c_void_p, SDL_TimerID, ctypes.c_uint32]]

SDL_AddTimer: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_AddTimer", SDL_TimerID, [ctypes.c_uint32, SDL_TimerCallback, ctypes.c_void_p], SDL_BINARY]

SDL_NSTimerCallback: typing.TypeAlias = SDL_FUNC_TYPE["SDL_NSTimerCallback", ctypes.c_uint64, [ctypes.c_void_p, SDL_TimerID, ctypes.c_uint64]]

SDL_AddTimerNS: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_AddTimerNS", SDL_TimerID, [ctypes.c_uint64, SDL_NSTimerCallback, ctypes.c_void_p], SDL_BINARY]
SDL_RemoveTimer: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_RemoveTimer", ctypes.c_bool, [SDL_TimerID], SDL_BINARY]


from .__init__ import ctypes, typing, \
    SDL_POINTER, SDL_TYPE, SDL_ENUM, SDL_FUNC_TYPE

from .SDL_rect import SDL_Rect
from .SDL_properties import SDL_PropertiesID
from .SDL_ttf import TTF_Font, TTF_Text, TTF_SubString
from .SDL_pixels import SDL_FColor

TTF_DrawCommand: typing.TypeAlias = SDL_TYPE["TTF_DrawCommand", SDL_ENUM]

TTF_DRAW_COMMAND_NOOP, TTF_DRAW_COMMAND_FILL, TTF_DRAW_COMMAND_COPY = range(3)

class TTF_FillOperation(ctypes.Structure):
    _fields_ = [
        ("cmd", TTF_DrawCommand),
        ("rect", SDL_Rect)
    ]

class TTF_CopyOperation(ctypes.Structure):
    _fields_ = [
        ("cmd", TTF_DrawCommand),
        ("text_offset", ctypes.c_int),
        ("glyph_font", SDL_POINTER[TTF_Font]),
        ("glyph_index", ctypes.c_uint32),
        ("src", SDL_Rect),
        ("dst", SDL_Rect),
        ("reserved", ctypes.c_void_p)
    ]

class TTF_DrawOperation(ctypes.Union):
    _fields_ = [
        ("cmd", TTF_DrawCommand),
        ("fill", TTF_FillOperation),
        ("copy", TTF_CopyOperation)
    ]

class TTF_TextLayout(ctypes.c_void_p):
    ...

class TTF_TextEngine(ctypes.Structure):
    _fields_ = [
        ("version", ctypes.c_uint32),
        ("userdata", ctypes.c_void_p),
        ("CreateText", SDL_FUNC_TYPE["TTF_TextEngine.CreateText", ctypes.c_bool, [ctypes.c_void_p, SDL_POINTER[TTF_Text]]]),
        ("DestroyText", SDL_FUNC_TYPE["TTF_TextEngine.DestroyText", None, [ctypes.c_void_p, SDL_POINTER[TTF_Text]]])
    ]

class TTF_TextData(ctypes.Structure):
    _fields_ = [
        ("font", SDL_POINTER[TTF_Font]),
        ("color", SDL_FColor),
        ("needs_layout_update", ctypes.c_bool),
        ("layout", SDL_POINTER[TTF_TextLayout]),
        ("x", ctypes.c_int),
        ("y", ctypes.c_int),
        ("w", ctypes.c_int),
        ("h", ctypes.c_int),
        ("num_ops", ctypes.c_int),
        ("ops", SDL_POINTER[TTF_DrawOperation]),
        ("num_clusters", ctypes.c_int),
        ("clusters", SDL_POINTER[TTF_SubString]),
        ("props", SDL_PropertiesID),
        ("needs_engine_update", ctypes.c_bool),
        ("engine", TTF_TextEngine),
        ("engine_text", ctypes.c_void_p)
    ]


from .__init__ import ctypes, typing, SDL_TYPE, SDL_ENUM

from .SDL_touch import SDL_TouchID
from .SDL_mouse import SDL_MouseID

SDL_PenID: typing.TypeAlias = SDL_TYPE["SDL_PenID", ctypes.c_uint32]

SDL_PEN_MOUSEID, SDL_PEN_TOUCHID = SDL_MouseID(-2), SDL_TouchID(-2)

SDL_PenInputFlags: typing.TypeAlias = SDL_TYPE["SDL_PenInputFlags", ctypes.c_uint32]

SDL_PEN_INPUT_DOWN: int = 1 << 0
SDL_PEN_INPUT_BUTTON_1: int = 1 << 1
SDL_PEN_INPUT_BUTTON_2: int = 1 << 2
SDL_PEN_INPUT_BUTTON_3: int = 1 << 3
SDL_PEN_INPUT_BUTTON_4: int = 1 << 4
SDL_PEN_INPUT_BUTTON_5: int = 1 << 5
SDL_PEN_INPUT_ERASER_TIP: int = 1 << 30

SDL_PenAxis: typing.TypeAlias = SDL_TYPE["SDL_PenAxis", SDL_ENUM]

SDL_PEN_AXIS_PRESSURE, SDL_PEN_AXIS_XTILT, SDL_PEN_AXIS_YTILT, SDL_PEN_AXIS_DISTANCE, SDL_PEN_AXIS_ROTATION, \
    SDL_PEN_AXIS_SLIDER, SDL_PEN_AXIS_TANGENTIAL_PRESSURE, SDL_PEN_AXIS_COUNT = range(8)


from .__init__ import ctypes, typing, abc, \
    SDL_POINTER, SDL_ENUM, SDL_FUNC, SDL_TYPE, SDL_BINARY

from .SDL_video import SDL_Window

SDL_MessageBoxFlags: typing.TypeAlias = SDL_TYPE["SDL_MessageBoxFlags", ctypes.c_uint32]

SDL_MESSAGEBOX_ERROR: int = 0x00000010
SDL_MESSAGEBOX_WARNING: int = 0x00000020
SDL_MESSAGEBOX_INFORMATION: int = 0x00000040
SDL_MESSAGEBOX_BUTTONS_LEFT_TO_RIGHT: int = 0x00000080
SDL_MESSAGEBOX_BUTTONS_RIGHT_TO_LEFT: int = 0x00000100

SDL_MessageBoxButtonFlags: typing.TypeAlias = SDL_TYPE["SDL_MessageBoxButtonFlags", ctypes.c_uint32]

SDL_MESSAGEBOX_BUTTON_RETURNKEY_DEFAULT: int = 0x00000001
SDL_MESSAGEBOX_BUTTON_ESCAPEKEY_DEFAULT: int = 0x00000002

class SDL_MessageBoxButtonData(ctypes.Structure):
    _fields_ = [
        ("flags", SDL_MessageBoxButtonFlags),
        ("buttonID", ctypes.c_int),
        ("text", ctypes.c_char_p)
    ]

class SDL_MessageBoxColor(ctypes.Structure):
    _fields_ = [
        ("r", ctypes.c_uint8),
        ("g", ctypes.c_uint8),
        ("b", ctypes.c_uint8)
    ]

SDL_MessageBoxColorType: typing.TypeAlias = SDL_TYPE["SDL_MessageBoxColorType", SDL_ENUM]

SDL_MESSAGEBOX_COLOR_BACKGROUND, SDL_MESSAGEBOX_COLOR_TEXT, SDL_MESSAGEBOX_COLOR_BUTTON_BORDER, \
    SDL_MESSAGEBOX_COLOR_BUTTON_BACKGROUND, SDL_MESSAGEBOX_COLOR_BUTTON_SELECTED, SDL_MESSAGEBOX_COLOR_COUNT = range(6)

class SDL_MessageBoxColorScheme(ctypes.Structure):
    _fields_ = [
        ("colors", SDL_MessageBoxColor * SDL_MESSAGEBOX_COLOR_COUNT)
    ]

class SDL_MessageBoxData(ctypes.Structure):
    _fields_ = [
        ("flags", SDL_MessageBoxFlags),
        ("window", SDL_POINTER[SDL_Window]),
        ("title", ctypes.c_char_p),
        ("message", ctypes.c_char_p),
        ("numbuttons", ctypes.c_int),
        ("buttons", SDL_POINTER[SDL_MessageBoxButtonData]),
        ("colorScheme", SDL_POINTER[SDL_MessageBoxColorScheme])
    ]

SDL_ShowMessageBox: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ShowMessageBox", ctypes.c_bool, [SDL_POINTER[SDL_MessageBoxData], SDL_POINTER[ctypes.c_int]], SDL_BINARY]
SDL_ShowSimpleMessageBox: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ShowSimpleMessageBox", ctypes.c_bool, [SDL_MessageBoxFlags, ctypes.c_char_p, ctypes.c_char_p, SDL_POINTER[SDL_Window]], SDL_BINARY]


from .__init__ import ctypes, typing, abc, \
    SDL_POINTER, SDL_ENUM, SDL_FUNC, SDL_TYPE, SDL_BINARY

from .SDL_guid import SDL_GUID
from .SDL_power import SDL_PowerState
from .SDL_sensor import SDL_SensorType
from .SDL_iostream import SDL_IOStream
from .SDL_properties import SDL_PropertiesID
from .SDL_joystick import SDL_Joystick, SDL_JoystickID, SDL_JoystickConnectionState, \
    SDL_PROP_JOYSTICK_CAP_MONO_LED_BOOLEAN, SDL_PROP_JOYSTICK_CAP_RGB_LED_BOOLEAN, SDL_PROP_JOYSTICK_CAP_PLAYER_LED_BOOLEAN, \
        SDL_PROP_JOYSTICK_CAP_RUMBLE_BOOLEAN, SDL_PROP_JOYSTICK_CAP_TRIGGER_RUMBLE_BOOLEAN

class SDL_Gamepad(ctypes.c_void_p):
    ...

SDL_GamepadType: typing.TypeAlias = SDL_TYPE["SDL_GamepadType", SDL_ENUM]

SDL_GAMEPAD_TYPE_UNKNOWN, SDL_GAMEPAD_TYPE_STANDARD, SDL_GAMEPAD_TYPE_XBOX360, SDL_GAMEPAD_TYPE_XBOXONE, SDL_GAMEPAD_TYPE_PS3, SDL_GAMEPAD_TYPE_PS4, SDL_GAMEPAD_TYPE_PS5, \
    SDL_GAMEPAD_TYPE_NINTENDO_SWITCH_PRO, SDL_GAMEPAD_TYPE_NINTENDO_SWITCH_JOYCON_LEFT, SDL_GAMEPAD_TYPE_NINTENDO_SWITCH_JOYCON_RIGHT, SDL_GAMEPAD_TYPE_NINTENDO_SWITCH_JOYCON_PAIR, \
        SDL_GAMEPAD_TYPE_COUNT = range(12)

SDL_GamepadButton: typing.TypeAlias = SDL_TYPE["SDL_GamepadButton", SDL_ENUM]

SDL_GAMEPAD_BUTTON_INVALID, SDL_GAMEPAD_BUTTON_SOUTH, SDL_GAMEPAD_BUTTON_EAST, SDL_GAMEPAD_BUTTON_WEST, SDL_GAMEPAD_BUTTON_NORTH, SDL_GAMEPAD_BUTTON_BACK, SDL_GAMEPAD_BUTTON_GUIDE, \
    SDL_GAMEPAD_BUTTON_START, SDL_GAMEPAD_BUTTON_LEFT_STICK, SDL_GAMEPAD_BUTTON_RIGHT_STICK, SDL_GAMEPAD_BUTTON_LEFT_SHOULDER, SDL_GAMEPAD_BUTTON_RIGHT_SHOULDER, SDL_GAMEPAD_BUTTON_DPAD_UP, \
        SDL_GAMEPAD_BUTTON_DPAD_DOWN, SDL_GAMEPAD_BUTTON_DPAD_LEFT, SDL_GAMEPAD_BUTTON_DPAD_RIGHT, SDL_GAMEPAD_BUTTON_MISC1, SDL_GAMEPAD_BUTTON_RIGHT_PADDLE1, SDL_GAMEPAD_BUTTON_LEFT_PADDLE1, \
            SDL_GAMEPAD_BUTTON_RIGHT_PADDLE2, SDL_GAMEPAD_BUTTON_LEFT_PADDLE2, SDL_GAMEPAD_BUTTON_TOUCHPAD, SDL_GAMEPAD_BUTTON_MISC2, SDL_GAMEPAD_BUTTON_MISC3, SDL_GAMEPAD_BUTTON_MISC4, \
                SDL_GAMEPAD_BUTTON_MISC5, SDL_GAMEPAD_BUTTON_MISC6, SDL_GAMEPAD_BUTTON_COUNT = range(-1, 27)

SDL_GamepadButtonLabel: typing.TypeAlias = SDL_TYPE["SDL_GamepadButtonLabel", SDL_ENUM]

SDL_GAMEPAD_BUTTON_LABEL_UNKNOWN, SDL_GAMEPAD_BUTTON_LABEL_A, SDL_GAMEPAD_BUTTON_LABEL_B, SDL_GAMEPAD_BUTTON_LABEL_X, SDL_GAMEPAD_BUTTON_LABEL_Y, \
    SDL_GAMEPAD_BUTTON_LABEL_CROSS, SDL_GAMEPAD_BUTTON_LABEL_CIRCLE, SDL_GAMEPAD_BUTTON_LABEL_SQUARE, SDL_GAMEPAD_BUTTON_LABEL_TRIANGLE = range(9)

SDL_GamepadAxis: typing.TypeAlias = SDL_TYPE["SDL_GamepadAxis", SDL_ENUM]

SDL_GAMEPAD_AXIS_INVALID, SDL_GAMEPAD_AXIS_LEFTX, SDL_GAMEPAD_AXIS_LEFTY, SDL_GAMEPAD_AXIS_RIGHTX, SDL_GAMEPAD_AXIS_RIGHTY, \
    SDL_GAMEPAD_AXIS_LEFT_TRIGGER, SDL_GAMEPAD_AXIS_RIGHT_TRIGGER, SDL_GAMEPAD_AXIS_COUNT = range(-1, 7)

SDL_GamepadBindingType: typing.TypeAlias = SDL_TYPE["SDL_GamepadBindingType", SDL_ENUM]

SDL_GAMEPAD_BINDTYPE_NONE, SDL_GAMEPAD_BINDTYPE_BUTTON, \
    SDL_GAMEPAD_BINDTYPE_AXIS, SDL_GAMEPAD_BINDTYPE_HAT = range(4)

class _SDL_GamepadBinding_axis(ctypes.Structure):
    _fields_ = [
        ("axis", SDL_GamepadAxis),
        ("axis_min", ctypes.c_int),
        ("axis_max", ctypes.c_int)
    ]

class _SDL_GamepadBinding_hat(ctypes.Structure):
    _fields_ = [
        ("hat", ctypes.c_int),
        ("hat_mask", ctypes.c_int)
    ]

class _SDL_GamepadBinding_input(ctypes.Union):
    _fields_ = [
        ("button", SDL_GamepadButton),
        ("axis", _SDL_GamepadBinding_axis),
        ("hat", _SDL_GamepadBinding_hat)
    ]

class _SDL_GamepadBinding_output(ctypes.Union):
    _fields_ = [
        ("button", SDL_GamepadButton),
        ("axis", _SDL_GamepadBinding_axis)
    ]

class SDL_GamepadBinding(ctypes.Structure):
    _fields_ = [
        ("input_type", SDL_GamepadBindingType),
        ("input", _SDL_GamepadBinding_input),
        ("output_type", SDL_GamepadBindingType),
        ("output", _SDL_GamepadBinding_output)
    ]

SDL_AddGamepadMapping: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_AddGamepadMapping", ctypes.c_int, [ctypes.c_char_p], SDL_BINARY]
SDL_AddGamepadMappingsFromIO: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_AddGamepadMappingsFromIO", ctypes.c_int, [SDL_POINTER[SDL_IOStream], ctypes.c_bool], SDL_BINARY]
SDL_AddGamepadMappingsFromFile: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_AddGamepadMappingsFromFile", ctypes.c_int, [ctypes.c_char_p], SDL_BINARY]
SDL_ReloadGamepadMappings: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ReloadGamepadMappings", ctypes.c_bool, [], SDL_BINARY]
SDL_GetGamepadMappings: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadMappings", SDL_POINTER[ctypes.c_char_p], [SDL_POINTER[ctypes.c_int]], SDL_BINARY]
SDL_GetGamepadMappingForGUID: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadMappingForGUID", ctypes.c_char_p, [SDL_GUID], SDL_BINARY]
SDL_GetGamepadMapping: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadMapping", ctypes.c_char_p, [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]
SDL_SetGamepadMapping: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetGamepadMapping", ctypes.c_bool, [SDL_JoystickID, ctypes.c_char_p], SDL_BINARY]
SDL_HasGamepad: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_HasGamepad", ctypes.c_bool, [], SDL_BINARY]
SDL_GetGamepads: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepads", SDL_POINTER[SDL_JoystickID], [SDL_POINTER[ctypes.c_int]], SDL_BINARY]
SDL_IsGamepad: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_IsGamepad", ctypes.c_bool, [SDL_JoystickID], SDL_BINARY]
SDL_GetGamepadNameForID: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadNameForID", ctypes.c_char_p, [SDL_JoystickID], SDL_BINARY]
SDL_GetGamepadPathForID: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadPathForID", ctypes.c_char_p, [SDL_JoystickID], SDL_BINARY]
SDL_GetGamepadPlayerIndexForID: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadPlayerIndexForID", ctypes.c_int, [SDL_JoystickID], SDL_BINARY]
SDL_GetGamepadGUIDForID: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadGUIDForID", SDL_GUID, [SDL_JoystickID], SDL_BINARY]
SDL_GetGamepadVendorForID: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadVendorForID", ctypes.c_uint16, [SDL_JoystickID], SDL_BINARY]
SDL_GetGamepadProductForID: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadProductForID", ctypes.c_uint16, [SDL_JoystickID], SDL_BINARY]
SDL_GetGamepadProductVersionForID: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadProductVersionForID", ctypes.c_uint16, [SDL_JoystickID], SDL_BINARY]
SDL_GetGamepadTypeForID: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadTypeForID", SDL_GamepadType, [SDL_JoystickID], SDL_BINARY]
SDL_GetRealGamepadTypeForID: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetRealGamepadTypeForID", SDL_GamepadType, [SDL_JoystickID], SDL_BINARY]
SDL_GetGamepadMappingForID: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadMappingForID", ctypes.c_char_p, [SDL_JoystickID], SDL_BINARY]
SDL_OpenGamepad: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_OpenGamepad", SDL_POINTER[SDL_Gamepad], [SDL_JoystickID], SDL_BINARY]
SDL_GetGamepadFromID: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadFromID", SDL_POINTER[SDL_Gamepad], [SDL_JoystickID], SDL_BINARY]
SDL_GetGamepadFromPlayerIndex: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadFromPlayerIndex", SDL_POINTER[SDL_Gamepad], [ctypes.c_int], SDL_BINARY]
SDL_GetGamepadProperties: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadProperties", SDL_PropertiesID, [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]

SDL_PROP_GAMEPAD_CAP_MONO_LED_BOOLEAN: bytes = SDL_PROP_JOYSTICK_CAP_MONO_LED_BOOLEAN
SDL_PROP_GAMEPAD_CAP_RGB_LED_BOOLEAN: bytes = SDL_PROP_JOYSTICK_CAP_RGB_LED_BOOLEAN
SDL_PROP_GAMEPAD_CAP_PLAYER_LED_BOOLEAN: bytes = SDL_PROP_JOYSTICK_CAP_PLAYER_LED_BOOLEAN
SDL_PROP_GAMEPAD_CAP_RUMBLE_BOOLEAN: bytes = SDL_PROP_JOYSTICK_CAP_RUMBLE_BOOLEAN
SDL_PROP_GAMEPAD_CAP_TRIGGER_RUMBLE_BOOLEAN: bytes = SDL_PROP_JOYSTICK_CAP_TRIGGER_RUMBLE_BOOLEAN

SDL_GetGamepadID: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadID", SDL_JoystickID, [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]
SDL_GetGamepadName: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadName", ctypes.c_char_p, [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]
SDL_GetGamepadPath: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadPath", ctypes.c_char_p, [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]
SDL_GetGamepadType: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadType", SDL_GamepadType, [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]
SDL_GetRealGamepadType: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetRealGamepadType", SDL_GamepadType, [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]
SDL_GetGamepadPlayerIndex: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadPlayerIndex", ctypes.c_int, [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]
SDL_SetGamepadPlayerIndex: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetGamepadPlayerIndex", ctypes.c_bool, [SDL_POINTER[SDL_Gamepad], ctypes.c_int], SDL_BINARY]
SDL_GetGamepadVendor: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadVendor", ctypes.c_uint16, [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]
SDL_GetGamepadProduct: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadProduct", ctypes.c_uint16, [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]
SDL_GetGamepadProductVersion: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadProductVersion", ctypes.c_uint16, [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]
SDL_GetGamepadFirmwareVersion: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadFirmwareVersion", ctypes.c_uint16, [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]
SDL_GetGamepadSerial: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadSerial", ctypes.c_char_p, [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]
SDL_GetGamepadSteamHandle: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadSteamHandle", ctypes.c_uint64, [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]
SDL_GetGamepadConnectionState: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadConnectionState", SDL_JoystickConnectionState, [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]
SDL_GetGamepadPowerInfo: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadPowerInfo", SDL_PowerState, [SDL_POINTER[SDL_Gamepad], SDL_POINTER[ctypes.c_int]], SDL_BINARY]
SDL_GamepadConnected: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GamepadConnected", ctypes.c_bool, [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]
SDL_GetGamepadJoystick: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadJoystick", SDL_POINTER[SDL_Joystick], [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]
SDL_SetGamepadEventsEnabled: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetGamepadEventsEnabled", None, [ctypes.c_bool], SDL_BINARY]
SDL_GamepadEventsEnabled: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GamepadEventsEnabled", ctypes.c_bool, [], SDL_BINARY]
SDL_GetGamepadBindings: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadBindings", SDL_POINTER[SDL_POINTER[SDL_GamepadBinding]], [SDL_POINTER[SDL_Gamepad], SDL_POINTER[ctypes.c_int]], SDL_BINARY]
SDL_UpdateGamepads: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_UpdateGamepads", None, [], SDL_BINARY]
SDL_GetGamepadTypeFromString: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadTypeFromString", SDL_GamepadType, [ctypes.c_char_p], SDL_BINARY]
SDL_GetGamepadStringForType: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadStringForType", ctypes.c_char_p, [SDL_GamepadType], SDL_BINARY]
SDL_GetGamepadAxisFromString: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadAxisFromString", SDL_GamepadAxis, [ctypes.c_char_p], SDL_BINARY]
SDL_GetGamepadStringForAxis: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadStringForAxis", ctypes.c_char_p, [SDL_GamepadAxis], SDL_BINARY]
SDL_GamepadHasAxis: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GamepadHasAxis", ctypes.c_bool, [SDL_POINTER[SDL_Gamepad], SDL_GamepadAxis], SDL_BINARY]
SDL_GetGamepadAxis: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadAxis", ctypes.c_int16, [SDL_POINTER[SDL_Gamepad], SDL_GamepadAxis], SDL_BINARY]
SDL_GetGamepadButtonFromString: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadButtonFromString", SDL_GamepadButton, [ctypes.c_char_p], SDL_BINARY]
SDL_GetGamepadStringForButton: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadStringForButton", ctypes.c_char_p, [SDL_GamepadButton], SDL_BINARY]
SDL_GamepadHasButton: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GamepadHasButton", ctypes.c_bool, [SDL_POINTER[SDL_Gamepad], SDL_GamepadButton], SDL_BINARY]
SDL_GetGamepadButton: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadButton", ctypes.c_bool, [SDL_POINTER[SDL_Gamepad], SDL_GamepadButton], SDL_BINARY]
SDL_GetGamepadButtonLabelForType: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadButtonLabelForType", SDL_GamepadButtonLabel, [SDL_GamepadType, SDL_GamepadButton], SDL_BINARY]
SDL_GetGamepadButtonLabel: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadButtonLabel", SDL_GamepadButtonLabel, [SDL_POINTER[SDL_Gamepad], SDL_GamepadButton], SDL_BINARY]
SDL_GetNumGamepadTouchpads: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetNumGamepadTouchpads", ctypes.c_int, [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]
SDL_GetNumGamepadTouchpadFingers: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetNumGamepadTouchpadFingers", ctypes.c_int, [SDL_POINTER[SDL_Gamepad], ctypes.c_int], SDL_BINARY]
SDL_GetGamepadTouchpadFinger: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadTouchpadFinger", ctypes.c_bool, [SDL_POINTER[SDL_Gamepad], ctypes.c_int, ctypes.c_int, SDL_POINTER[ctypes.c_bool], SDL_POINTER[ctypes.c_float], SDL_POINTER[ctypes.c_float], SDL_POINTER[ctypes.c_float]], SDL_BINARY]
SDL_GamepadHasSensor: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GamepadHasSensor", ctypes.c_bool, [SDL_POINTER[SDL_Gamepad], SDL_SensorType], SDL_BINARY]
SDL_SetGamepadSensorEnabled: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetGamepadSensorEnabled", ctypes.c_bool, [SDL_POINTER[SDL_Gamepad], SDL_SensorType, ctypes.c_bool], SDL_BINARY]
SDL_GamepadSensorEnabled: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GamepadSensorEnabled", ctypes.c_bool, [SDL_POINTER[SDL_Gamepad], SDL_SensorType], SDL_BINARY]
SDL_GetGamepadSensorDataRate: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadSensorDataRate", ctypes.c_float, [SDL_POINTER[SDL_Gamepad], SDL_SensorType], SDL_BINARY]
SDL_GetGamepadSensorData: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadSensorData", ctypes.c_bool, [SDL_POINTER[SDL_Gamepad], SDL_SensorType, SDL_POINTER[ctypes.c_float], ctypes.c_int], SDL_BINARY]
SDL_RumbleGamepad: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_RumbleGamepad", ctypes.c_bool, [SDL_POINTER[SDL_Gamepad], ctypes.c_uint16, ctypes.c_uint16, ctypes.c_uint32], SDL_BINARY]
SDL_RumbleGamepadTriggers: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_RumbleGamepadTriggers", ctypes.c_bool, [SDL_POINTER[SDL_Gamepad], ctypes.c_uint16, ctypes.c_uint16, ctypes.c_uint32], SDL_BINARY]
SDL_SetGamepadLED: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetGamepadLED", ctypes.c_bool, [SDL_POINTER[SDL_Gamepad], ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8], SDL_BINARY]
SDL_SendGamepadEffect: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SendGamepadEffect", ctypes.c_bool, [SDL_POINTER[SDL_Gamepad], ctypes.c_void_p, ctypes.c_int], SDL_BINARY]
SDL_CloseGamepad: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_CloseGamepad", None, [SDL_POINTER[SDL_Gamepad]], SDL_BINARY]

SDL_GetGamepadAppleSFSymbolsNameForButton: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadAppleSFSymbolsNameForButton", ctypes.c_char_p, [SDL_POINTER[SDL_Gamepad], SDL_GamepadButton], SDL_BINARY]
SDL_GetGamepadAppleSFSymbolsNameForAxis: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetGamepadAppleSFSymbolsNameForAxis", ctypes.c_char_p, [SDL_POINTER[SDL_Gamepad], SDL_GamepadAxis], SDL_BINARY]


from .__init__ import ctypes, typing, abc, \
    SDL_POINTER, SDL_ENUM, SDL_FUNC, SDL_TYPE, SDL_BINARY

from .SDL_stdinc import SDL_FOURCC
from .SDL_endian import SDL_BYTEORDER, SDL_BIG_ENDIAN

SDL_ALPHA_OPAQUE, SDL_ALPHA_TRANSPARENT_FLOAT = 255, 1.0
SDL_ALPHA_TRANSPARENT, SDL_ALPHA_OPAQUE_FLOAT = 0, 0.0

SDL_PixelType: typing.TypeAlias = SDL_TYPE["SDL_PixelType", SDL_ENUM]

SDL_PIXELTYPE_UNKNOWN, SDL_PIXELTYPE_INDEX1, SDL_PIXELTYPE_INDEX4, SDL_PIXELTYPE_INDEX8, SDL_PIXELTYPE_PACKED8, SDL_PIXELTYPE_PACKED16, SDL_PIXELTYPE_PACKED32, \
    SDL_PIXELTYPE_ARRAYU8, SDL_PIXELTYPE_ARRAYU16, SDL_PIXELTYPE_ARRAYU32, SDL_PIXELTYPE_ARRAYF16, SDL_PIXELTYPE_ARRAYF32, SDL_PIXELTYPE_INDEX2 = range(13)

SDL_BitmapOrder: typing.TypeAlias = SDL_TYPE["SDL_BitmapOrder", SDL_ENUM]

SDL_BITMAPORDER_NONE, SDL_BITMAPORDER_4321, SDL_BITMAPORDER_1234 = range(3)

SDL_PackedOrder: typing.TypeAlias = SDL_TYPE["SDL_PackedOrder", SDL_ENUM]

SDL_PACKEDORDER_NONE, SDL_PACKEDORDER_XRGB, SDL_PACKEDORDER_RGBX, SDL_PACKEDORDER_ARGB, SDL_PACKEDORDER_RGBA, \
    SDL_PACKEDORDER_XBGR, SDL_PACKEDORDER_BGRX, SDL_PACKEDORDER_ABGR, SDL_PACKEDORDER_BGRA = range(9)

SDL_ArrayOrder: typing.TypeAlias = SDL_TYPE["SDL_ArrayOrder", SDL_ENUM]

SDL_ARRAYORDER_NONE, SDL_ARRAYORDER_RGB, SDL_ARRAYORDER_RGBA, SDL_ARRAYORDER_ARGB, \
    SDL_ARRAYORDER_BGR, SDL_ARRAYORDER_BGRA, SDL_ARRAYORDER_ABGR = range(7)

SDL_PackedLayout: typing.TypeAlias = SDL_TYPE["SDL_PackedLayout", SDL_ENUM]

SDL_PACKEDLAYOUT_NONE, SDL_PACKEDLAYOUT_332, SDL_PACKEDLAYOUT_4444, SDL_PACKEDLAYOUT_1555, SDL_PACKEDLAYOUT_5551, SDL_PACKEDLAYOUT_565, \
    SDL_PACKEDLAYOUT_8888, SDL_PACKEDLAYOUT_2101010, SDL_PACKEDLAYOUT_1010102 = range(9)

def SDL_DEFINE_PIXELFORMAT(ptype: int, order: int, layout: int, bits: int, pbytes: int) -> int:
    return (1 << 28) | (ptype << 24) | (order << 20) | (layout << 16) | (bits << 8) | (pbytes << 0)

SDL_DEFINE_PIXELFOURCC: abc.Callable[[int, int, int, int], int] = SDL_FOURCC
SDL_ISPIXELFORMAT_FOURCC: abc.Callable[..., bool] = lambda format: format and SDL_PIXELFLAG(format) != 1

SDL_PIXELFLAG: abc.Callable[..., int] = lambda x: (x >> 28) & 0x0F
SDL_PIXELTYPE: abc.Callable[..., int] = lambda x: (x >> 24) & 0x0F
SDL_PIXELORDER: abc.Callable[..., int] = lambda x: (x >> 20) & 0x0F
SDL_PIXELLAYOUT: abc.Callable[..., int] = lambda x: (x >> 16) & 0x0F
SDL_BITSPERPIXEL: abc.Callable[..., int] = lambda x: (x >> 8) & 0xFF

def SDL_BYTESPERPIXEL(x: int) -> int:
    if not SDL_ISPIXELFORMAT_FOURCC(x): return (x >> 0) & 0xFF
    else: return 2 if x in (SDL_PIXELFORMAT_YUY2, SDL_PIXELFORMAT_UYVY, SDL_PIXELFORMAT_YVYU, SDL_PIXELFORMAT_P010) else 1

def SDL_ISPIXELFORMAT_INDEXED(format: int) -> bool:
    return not SDL_ISPIXELFORMAT_FOURCC(format) and (SDL_PIXELTYPE(format) in [SDL_PIXELTYPE_INDEX1, SDL_PIXELTYPE_INDEX2, SDL_PIXELTYPE_INDEX4, SDL_PIXELTYPE_INDEX8])

def SDL_ISPIXELFORMAT_PACKED(format: int) -> bool:
    return not SDL_ISPIXELFORMAT_FOURCC(format) and (SDL_PIXELTYPE(format) in [SDL_PIXELTYPE_PACKED8, SDL_PIXELTYPE_PACKED16, SDL_PIXELTYPE_PACKED32])

def SDL_ISPIXELFORMAT_ARRAY(format: int) -> bool:
    return not SDL_ISPIXELFORMAT_FOURCC(format) and SDL_PIXELTYPE(format) in [SDL_PIXELTYPE_ARRAYU8, SDL_PIXELTYPE_ARRAYU16, SDL_PIXELTYPE_ARRAYU32, SDL_PIXELTYPE_ARRAYF16, SDL_PIXELTYPE_ARRAYF32]

def SDL_ISPIXELFORMAT_10BIT(format: int) -> bool:
    return not SDL_ISPIXELFORMAT_FOURCC(format) and (SDL_PIXELTYPE(format) == SDL_PIXELTYPE_PACKED32 and SDL_PIXELLAYOUT(format) == SDL_PACKEDLAYOUT_2101010)

def SDL_ISPIXELFORMAT_FLOAT(format: int) -> bool:
    return not SDL_ISPIXELFORMAT_FOURCC(format) and (SDL_PIXELTYPE(format) == SDL_PIXELTYPE_ARRAYF16 or SDL_PIXELTYPE(format) == SDL_PIXELTYPE_ARRAYF32)

def SDL_ISPIXELFORMAT_ALPHA(format: int) -> bool:
    return SDL_ISPIXELFORMAT_PACKED(format) and (SDL_PIXELORDER(format) in [SDL_PACKEDORDER_ARGB, SDL_PACKEDORDER_RGBA, SDL_PACKEDORDER_ABGR, SDL_PACKEDORDER_BGRA]) \
        or SDL_ISPIXELFORMAT_ARRAY(format) and (SDL_PIXELORDER(format) in [SDL_ARRAYORDER_ARGB, SDL_ARRAYORDER_RGBA, SDL_ARRAYORDER_ABGR, SDL_ARRAYORDER_BGRA])

SDL_PixelFormat: typing.TypeAlias = SDL_TYPE["SDL_PixelFormat", SDL_ENUM]

SDL_PIXELFORMAT_UNKNOWN: int = 0x00000000
SDL_PIXELFORMAT_INDEX1LSB: int = 0x11100100
SDL_PIXELFORMAT_INDEX1MSB: int = 0x11200100
SDL_PIXELFORMAT_INDEX2LSB: int = 0x1c100200
SDL_PIXELFORMAT_INDEX2MSB: int = 0x1c200200
SDL_PIXELFORMAT_INDEX4LSB: int = 0x12100400
SDL_PIXELFORMAT_INDEX4MSB: int = 0x12200400
SDL_PIXELFORMAT_INDEX8: int = 0x13000801
SDL_PIXELFORMAT_RGB332: int = 0x14110801
SDL_PIXELFORMAT_XRGB4444: int = 0x15120c02
SDL_PIXELFORMAT_XBGR4444: int = 0x15520c02
SDL_PIXELFORMAT_XRGB1555: int = 0x15130f02
SDL_PIXELFORMAT_XBGR1555: int = 0x15530f02
SDL_PIXELFORMAT_ARGB4444: int = 0x15321002
SDL_PIXELFORMAT_RGBA4444: int = 0x15421002
SDL_PIXELFORMAT_ABGR4444: int = 0x15721002
SDL_PIXELFORMAT_BGRA4444: int = 0x15821002
SDL_PIXELFORMAT_ARGB1555: int = 0x15331002
SDL_PIXELFORMAT_RGBA5551: int = 0x15441002
SDL_PIXELFORMAT_ABGR1555: int = 0x15731002
SDL_PIXELFORMAT_BGRA5551: int = 0x15841002
SDL_PIXELFORMAT_RGB565: int = 0x15151002
SDL_PIXELFORMAT_BGR565: int = 0x15551002
SDL_PIXELFORMAT_RGB24: int = 0x17101803
SDL_PIXELFORMAT_BGR24: int = 0x17401803
SDL_PIXELFORMAT_XRGB8888: int = 0x16161804
SDL_PIXELFORMAT_RGBX8888: int = 0x16261804
SDL_PIXELFORMAT_XBGR8888: int = 0x16561804
SDL_PIXELFORMAT_BGRX8888: int = 0x16661804
SDL_PIXELFORMAT_ARGB8888: int = 0x16362004
SDL_PIXELFORMAT_RGBA8888: int = 0x16462004
SDL_PIXELFORMAT_ABGR8888: int = 0x16762004
SDL_PIXELFORMAT_BGRA8888: int = 0x16862004
SDL_PIXELFORMAT_XRGB2101010: int = 0x16172004
SDL_PIXELFORMAT_XBGR2101010: int = 0x16572004
SDL_PIXELFORMAT_ARGB2101010: int = 0x16372004
SDL_PIXELFORMAT_ABGR2101010: int = 0x16772004
SDL_PIXELFORMAT_RGB48: int = 0x18103006
SDL_PIXELFORMAT_BGR48: int = 0x18403006
SDL_PIXELFORMAT_RGBA64: int = 0x18204008
SDL_PIXELFORMAT_ARGB64: int = 0x18304008
SDL_PIXELFORMAT_BGRA64: int = 0x18504008
SDL_PIXELFORMAT_ABGR64: int = 0x18604008
SDL_PIXELFORMAT_RGB48_FLOAT: int = 0x1a103006
SDL_PIXELFORMAT_BGR48_FLOAT: int = 0x1a403006
SDL_PIXELFORMAT_RGBA64_FLOAT: int = 0x1a204008
SDL_PIXELFORMAT_ARGB64_FLOAT: int = 0x1a304008
SDL_PIXELFORMAT_BGRA64_FLOAT: int = 0x1a504008
SDL_PIXELFORMAT_ABGR64_FLOAT: int = 0x1a604008
SDL_PIXELFORMAT_RGB96_FLOAT: int = 0x1b10600c
SDL_PIXELFORMAT_BGR96_FLOAT: int = 0x1b40600c
SDL_PIXELFORMAT_RGBA128_FLOAT: int = 0x1b208010
SDL_PIXELFORMAT_ARGB128_FLOAT: int = 0x1b308010
SDL_PIXELFORMAT_BGRA128_FLOAT: int = 0x1b508010
SDL_PIXELFORMAT_ABGR128_FLOAT: int = 0x1b608010
SDL_PIXELFORMAT_YV12: int = 0x32315659
SDL_PIXELFORMAT_IYUV: int = 0x56555949
SDL_PIXELFORMAT_YUY2: int = 0x32595559
SDL_PIXELFORMAT_UYVY: int = 0x59565955
SDL_PIXELFORMAT_YVYU: int = 0x55595659
SDL_PIXELFORMAT_NV12: int = 0x3231564e
SDL_PIXELFORMAT_NV21: int = 0x3132564e
SDL_PIXELFORMAT_P010: int = 0x30313050
SDL_PIXELFORMAT_EXTERNAL_OES: int = 0x2053454f
SDL_PIXELFORMAT_MJPG: int = 0x47504a4d

if SDL_BYTEORDER == SDL_BIG_ENDIAN:
    SDL_PIXELFORMAT_RGBA32, SDL_PIXELFORMAT_ARGB32, SDL_PIXELFORMAT_BGRA32, SDL_PIXELFORMAT_ABGR32, SDL_PIXELFORMAT_RGBX32, SDL_PIXELFORMAT_XRGB32, SDL_PIXELFORMAT_BGRX32, SDL_PIXELFORMAT_XBGR32 = \
        SDL_PIXELFORMAT_RGBA8888, SDL_PIXELFORMAT_ARGB8888, SDL_PIXELFORMAT_BGRA8888, SDL_PIXELFORMAT_ABGR8888, SDL_PIXELFORMAT_RGBX8888, SDL_PIXELFORMAT_XRGB8888, SDL_PIXELFORMAT_BGRX8888, SDL_PIXELFORMAT_XBGR8888

else:
    SDL_PIXELFORMAT_RGBA32, SDL_PIXELFORMAT_ARGB32, SDL_PIXELFORMAT_BGRA32, SDL_PIXELFORMAT_ABGR32, SDL_PIXELFORMAT_RGBX32, SDL_PIXELFORMAT_XRGB32, SDL_PIXELFORMAT_BGRX32, SDL_PIXELFORMAT_XBGR32 = \
        SDL_PIXELFORMAT_ABGR8888, SDL_PIXELFORMAT_BGRA8888, SDL_PIXELFORMAT_ARGB8888, SDL_PIXELFORMAT_RGBA8888, SDL_PIXELFORMAT_XBGR8888, SDL_PIXELFORMAT_BGRX8888, SDL_PIXELFORMAT_XRGB8888, SDL_PIXELFORMAT_RGBX8888

SDL_ColorType: typing.TypeAlias = SDL_TYPE["SDL_ColorType", SDL_ENUM]

SDL_COLOR_TYPE_UNKNOWN, SDL_COLOR_TYPE_RGB, SDL_COLOR_TYPE_YCBCR = range(3)

SDL_ColorRange: typing.TypeAlias = SDL_TYPE["SDL_ColorRange", SDL_ENUM]

SDL_COLOR_RANGE_UNKOWN, SDL_COLOR_RANGE_LIMITED, SDL_COLOR_RANGE_FULL = range(3)

SDL_ColorPrimaries: typing.TypeAlias = SDL_TYPE["SDL_ColorPrimaries", SDL_ENUM]

SDL_COLOR_PRIMARIES_UNKNOWN, SDL_COLOR_PRIMARIES_BT709, SDL_COLOR_PRIMARIES_UNSPECIFIED, SDL_COLOR_PRIMARIES_BT470M, SDL_COLOR_PRIMARIES_BT470BG, \
    SDL_COLOR_PRIMARIES_BT601, SDL_COLOR_PRIMARIES_SMPTE240, SDL_COLOR_PRIMARIES_GENERIC_FILM, SDL_COLOR_PRIMARIES_BT2020, SDL_COLOR_PRIMARIES_XYZ , \
        SDL_COLOR_PRIMARIES_SMPTE431 , SDL_COLOR_PRIMARIES_SMPTE432 = range(12)

SDL_COLOR_PRIMARIES_EBU3213, SDL_COLOR_PRIMARIES_CUSTOM = 22, 31

SDL_TransferCharacteristics: typing.TypeAlias = SDL_TYPE["SDL_TransferCharacteristics", SDL_ENUM]

SDL_TRANSFER_CHARACTERISTICS_UNKNOWN, SDL_TRANSFER_CHARACTERISTICS_BT709, SDL_TRANSFER_CHARACTERISTICS_UNSPECIFIED, SDL_TRANSFER_CHARACTERISTICS_GAMMA22, \
    SDL_TRANSFER_CHARACTERISTICS_GAMMA28, SDL_TRANSFER_CHARACTERISTICS_BT601, SDL_TRANSFER_CHARACTERISTICS_SMPTE240, SDL_TRANSFER_CHARACTERISTICS_LINEAR, SDL_TRANSFER_CHARACTERISTICS_LOG100, \
        SDL_TRANSFER_CHARACTERISTICS_LOG100_SQRT10, SDL_TRANSFER_CHARACTERISTICS_IEC61966, SDL_TRANSFER_CHARACTERISTICS_BT1361, SDL_TRANSFER_CHARACTERISTICS_SRGB, SDL_TRANSFER_CHARACTERISTICS_BT2020_10BIT, \
            SDL_TRANSFER_CHARACTERISTICS_BT2020_12BIT, SDL_TRANSFER_CHARACTERISTICS_PQ, SDL_TRANSFER_CHARACTERISTICS_SMPTE428, SDL_TRANSFER_CHARACTERISTICS_HLG = range(18)

SDL_TRANSFER_CHARACTERISTICS_CUSTOM: int = 31

SDL_MatrixCoefficients: typing.TypeAlias = SDL_TYPE["SDL_MatrixCoefficients", SDL_ENUM]

SDL_MATRIX_COEFFICIENTS_IDENTITY, SDL_MATRIX_COEFFICIENTS_BT709, SDL_MATRIX_COEFFICIENTS_UNSPECIFIED, SDL_MATRIX_COEFFICIENTS_FCC, \
    SDL_MATRIX_COEFFICIENTS_BT470BG, SDL_MATRIX_COEFFICIENTS_BT601, SDL_MATRIX_COEFFICIENTS_SMPTE240, SDL_MATRIX_COEFFICIENTS_YCGCO, \
        SDL_MATRIX_COEFFICIENTS_BT2020_NCL, SDL_MATRIX_COEFFICIENTS_BT2020_CL, SDL_MATRIX_COEFFICIENTS_SMPTE2085, SDL_MATRIX_COEFFICIENTS_CHROMA_DERIVED_NCL, \
            SDL_MATRIX_COEFFICIENTS_CHROMA_DERIVED_CL, SDL_MATRIX_COEFFICIENTS_ICTCP = range(14)

SDL_MATRIX_COEFFICIENTS_CUSTOM: int = 31

SDL_ChromaLocation: typing.TypeAlias = SDL_TYPE["SDL_ChromaLocation", SDL_ENUM]

SDL_CHROMA_LOCATION_NONE, SDL_CHROMA_LOCATION_LEFT, \
    SDL_CHROMA_LOCATION_CENTER, SDL_CHROMA_LOCATION_TOPLEFT = range(4)

SDL_Colorspace: typing.TypeAlias = SDL_TYPE["SDL_Colorspace", SDL_ENUM]

SDL_COLORSPACE_UNKNOWN: int = 0x00000000
SDL_COLORSPACE_SRGB: int = 0x120005a0
SDL_COLORSPACE_SRGB_LINEAR: int = 0x12000500
SDL_COLORSPACE_HDR10: int = 0x12002600
SDL_COLORSPACE_JPEG: int = 0x220004c6
SDL_COLORSPACE_BT601_LIMITED: int = 0x211018c6
SDL_COLORSPACE_BT601_FULL: int = 0x221018c6
SDL_COLORSPACE_BT709_LIMITED: int = 0x21100421
SDL_COLORSPACE_BT709_FULL: int = 0x22100421
SDL_COLORSPACE_BT2020_LIMITED: int = 0x21102609
SDL_COLORSPACE_BT2020_FULL: int = 0x22102609

SDL_COLORSPACE_RGB_DEFAULT, SDL_COLORSPACE_YUV_DEFAULT = \
    SDL_COLORSPACE_SRGB, SDL_COLORSPACE_JPEG

class SDL_Color(ctypes.Structure):
    _fields_ = [
        ("r", ctypes.c_uint8),
        ("g", ctypes.c_uint8),
        ("b", ctypes.c_uint8),
        ("a", ctypes.c_uint8)
    ]

class SDL_FColor(ctypes.Structure):
    _fields_ = [
        ("r", ctypes.c_float),
        ("g", ctypes.c_float),
        ("b", ctypes.c_float),
        ("a", ctypes.c_float)
    ]

class SDL_Palette(ctypes.Structure):
    _fields_ = [
        ("ncolors", ctypes.c_int),
        ("colors", SDL_POINTER[SDL_Color]),
        ("version", ctypes.c_uint32),
        ("refcount", ctypes.c_int)
    ]

class SDL_PixelFormatDetails(ctypes.Structure):
    _fields_ = [
        ("format", SDL_PixelFormat),
        ("bits_per_pixel", ctypes.c_uint8),
        ("bytes_per_pixel", ctypes.c_uint8),
        ("padding", ctypes.c_uint8 * 2),
        ("Rmask", ctypes.c_uint32),
        ("Gmask", ctypes.c_uint32),
        ("Bmask", ctypes.c_uint32),
        ("Amask", ctypes.c_uint32),
        ("Rbits", ctypes.c_uint8),
        ("Gbits", ctypes.c_uint8),
        ("Bbits", ctypes.c_uint8),
        ("Abits", ctypes.c_uint8),
        ("Rshift", ctypes.c_uint8),
        ("Gshift", ctypes.c_uint8),
        ("Bshift", ctypes.c_uint8),
        ("Ashift", ctypes.c_uint8)
    ]

SDL_GetPixelFormatName: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetPixelFormatName", ctypes.c_char_p, [SDL_PixelFormat], SDL_BINARY]
SDL_GetMasksForPixelFormat: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetMasksForPixelFormat", ctypes.c_bool, [SDL_PixelFormat, SDL_POINTER[ctypes.c_int], SDL_POINTER[ctypes.c_uint32], SDL_POINTER[ctypes.c_uint32], SDL_POINTER[ctypes.c_uint32], SDL_POINTER[ctypes.c_uint32]], SDL_BINARY]
SDL_GetPixelFormatForMasks: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetPixelFormatForMasks", SDL_PixelFormat, [ctypes.c_int, ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint32], SDL_BINARY]
SDL_GetPixelFormatDetails: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetPixelFormatDetails", SDL_POINTER[SDL_PixelFormatDetails], [SDL_PixelFormat], SDL_BINARY]
SDL_CreatePalette: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_CreatePalette", SDL_POINTER[SDL_Palette], [ctypes.c_int], SDL_BINARY]
SDL_SetPaletteColors: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_SetPaletteColors", ctypes.c_bool, [SDL_POINTER[SDL_Palette], SDL_POINTER[SDL_Color], ctypes.c_int, ctypes.c_int], SDL_BINARY]
SDL_DestroyPalette: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_DestroyPalette", None, [SDL_POINTER[SDL_Palette]], SDL_BINARY]
SDL_MapRGB: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_MapRGB", ctypes.c_uint32, [SDL_POINTER[SDL_PixelFormatDetails], SDL_POINTER[SDL_Palette], ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8], SDL_BINARY]
SDL_MapRGBA: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_MapRGBA", ctypes.c_uint32, [SDL_POINTER[SDL_PixelFormatDetails], SDL_POINTER[SDL_Palette], ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8], SDL_BINARY]
SDL_GetRGB: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetRGB", None, [ctypes.c_uint32, SDL_POINTER[SDL_PixelFormatDetails], SDL_POINTER[SDL_Palette], SDL_POINTER[ctypes.c_uint8], SDL_POINTER[ctypes.c_uint8], SDL_POINTER[ctypes.c_uint8]], SDL_BINARY]
SDL_GetRGBA: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetRGBA", None, [ctypes.c_uint32, SDL_POINTER[SDL_PixelFormatDetails], SDL_POINTER[SDL_Palette], SDL_POINTER[ctypes.c_uint8], SDL_POINTER[ctypes.c_uint8], SDL_POINTER[ctypes.c_uint8], SDL_POINTER[ctypes.c_uint8]], SDL_BINARY]


from .__init__ import *

def SDL_MostSignificantBitIndex32(x: int) -> int:
    if x == 0: return -1
    b: list[int] = [0x2, 0xC, 0xF0, 0xFF00, 0xFFFF0000]
    S: list[int] = [1, 2, 4, 8, 16]
    msbIndex: int = 0

    for i in range(4, -1, -1):
        if x & b[i]:
            x >>= S[i]
            msbIndex |= S[i]

    return msbIndex

def SDL_HasExactlyOneBitSet32(x: int) -> bool:
    if x and not (x & (x - 1)): return True
    else: return False


from .__init__ import ctypes, typing, abc, \
    SDL_POINTER, SDL_FUNC_TYPE, SDL_FUNC, SDL_BINARY

from .SDL_properties import SDL_PropertiesID
from .SDL_filesystem import SDL_GlobFlags, \
    SDL_EnumerateDirectoryCallback, SDL_PathInfo

class SDL_StorageInterface(ctypes.Structure):
    _fields_ = [
        ("version", ctypes.c_uint32),
        ("close", SDL_FUNC_TYPE["SDL_StorageInterface.close", ctypes.c_bool, [ctypes.c_void_p]]),
        ("ready", SDL_FUNC_TYPE["SDL_StorageInterface.ready", ctypes.c_bool, [ctypes.c_void_p]]),
        ("enumerate", SDL_FUNC_TYPE["SDL_StorageInterface.enumerate", ctypes.c_bool, [ctypes.c_void_p, ctypes.c_char_p, SDL_EnumerateDirectoryCallback, ctypes.c_void_p]]),
        ("info", SDL_FUNC_TYPE["SDL_StorageInterface.info", ctypes.c_bool, [ctypes.c_void_p, ctypes.c_char_p, SDL_POINTER[SDL_PathInfo]]]),
        ("read_file", SDL_FUNC_TYPE["SDL_StorageInterface.read_file", ctypes.c_bool, [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_void_p, ctypes.c_uint64]]),
        ("write_file", SDL_FUNC_TYPE["SDL_StorageInterface.write_file", ctypes.c_bool, [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_void_p, ctypes.c_uint64]]),
        ("mkdir", SDL_FUNC_TYPE["SDL_StorageInterface.mkdir", ctypes.c_bool, [ctypes.c_void_p, ctypes.c_char_p]]),
        ("remove", SDL_FUNC_TYPE["SDL_StorageInterface.remove", ctypes.c_bool, [ctypes.c_void_p, ctypes.c_char_p]]),
        ("rename", SDL_FUNC_TYPE["SDL_StorageInterface.rename", ctypes.c_bool, [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p]]),
        ("copy", SDL_FUNC_TYPE["SDL_StorageInterface.copy", ctypes.c_bool, [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p]]),
        ("space_remaining", SDL_FUNC_TYPE["SDL_StorageInterface.space_remaining", ctypes.c_uint64, [ctypes.c_void_p]])
    ]

class SDL_Storage(ctypes.c_void_p):
    ...

SDL_OpenTitleStorage: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_OpenTitleStorage", SDL_POINTER[SDL_Storage], [ctypes.c_char_p, SDL_PropertiesID], SDL_BINARY]
SDL_OpenUserStorage: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_OpenUserStorage", SDL_POINTER[SDL_Storage], [ctypes.c_char_p, ctypes.c_char_p, SDL_PropertiesID], SDL_BINARY]
SDL_OpenFileStorage: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_OpenFileStorage", SDL_POINTER[SDL_Storage], [ctypes.c_char_p], SDL_BINARY]
SDL_OpenStorage: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_OpenStorage", SDL_POINTER[SDL_Storage], [SDL_POINTER[SDL_StorageInterface], ctypes.c_void_p], SDL_BINARY]
SDL_CloseStorage: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_CloseStorage", ctypes.c_bool, [SDL_POINTER[SDL_Storage]], SDL_BINARY]
SDL_StorageReady: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_StorageReady", ctypes.c_bool, [SDL_POINTER[SDL_Storage]], SDL_BINARY]
SDL_GetStorageFileSize: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetStorageFileSize", ctypes.c_bool, [SDL_POINTER[SDL_Storage], ctypes.c_char_p, SDL_POINTER[ctypes.c_uint64]], SDL_BINARY]
SDL_ReadStorageFile: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_ReadStorageFile", ctypes.c_bool, [SDL_POINTER[SDL_Storage], ctypes.c_char_p, ctypes.c_void_p, ctypes.c_uint64], SDL_BINARY]
SDL_WriteStorageFile: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_WriteStorageFile", ctypes.c_bool, [SDL_POINTER[SDL_Storage], ctypes.c_char_p, ctypes.c_void_p, ctypes.c_uint64], SDL_BINARY]
SDL_CreateStorageDirectory: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_CreateStorageDirectory", ctypes.c_bool, [SDL_POINTER[SDL_Storage], ctypes.c_char_p], SDL_BINARY]
SDL_EnumerateStorageDirectory: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_EnumerateStorageDirectory", ctypes.c_bool, [SDL_POINTER[SDL_Storage], ctypes.c_char_p, SDL_EnumerateDirectoryCallback, ctypes.c_void_p], SDL_BINARY]
SDL_RemoveStoragePath: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_RemoveStoragePath", ctypes.c_bool, [SDL_POINTER[SDL_Storage], ctypes.c_char_p], SDL_BINARY]
SDL_RenameStoragePath: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_RenameStoragePath", ctypes.c_bool, [SDL_POINTER[SDL_Storage], ctypes.c_char_p, ctypes.c_char_p], SDL_BINARY]
SDL_CopyStorageFile: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_CopyStorageFile", ctypes.c_bool, [SDL_POINTER[SDL_Storage], ctypes.c_char_p, ctypes.c_char_p], SDL_BINARY]
SDL_GetStoragePathInfo: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetStoragePathInfo", ctypes.c_bool, [SDL_POINTER[SDL_Storage], ctypes.c_char_p, SDL_POINTER[SDL_PathInfo]], SDL_BINARY]
SDL_GetStorageSpaceRemaining: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GetStorageSpaceRemaining", ctypes.c_uint64, [SDL_POINTER[SDL_Storage]], SDL_BINARY]
SDL_GlobStorageDirectory: abc.Callable[..., typing.Any] = SDL_FUNC["SDL_GlobStorageDirectory", SDL_POINTER[ctypes.c_char_p], [SDL_POINTER[SDL_Storage], ctypes.c_char_p, ctypes.c_char_p, SDL_GlobFlags, SDL_POINTER[ctypes.c_int]], SDL_BINARY]